/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @headerfile  common.h 
 *
 * @brief       Defines common interfaces for libmodal-cv
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2024
 */

#ifndef MODALCV_COMMON_H
#define MODALCV_COMMON_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

/**
 * ERROR CODES
 */
#define MCV_SUCCESS				0
#define MCV_EFAIL				1
#define MCV_EUNALIGNPARAM		2
#define MCV_EBADPARAM			3
#define MCV_EMISSINGTASK		4
#define MCV_EINVALSTATE			5
#define MCV_ENORES				6
#define MCV_EUNSUPPORTED		7
#define MCV_ECLOCKEXCEEDSLIMIT	8
#define MCV_EFATAL				9

/**
 * @brief Prints the corresponding error argument 
 * 
 * @param err The error type as specified above
 * @return 0 if successful, -1 otherwise
 */
int mcv_print_error(int err);

/**
 * @brief Globally enables/disables debug prints for libmodal-cv
 * 
 * @param en 1 to enable, 0 to disable
 */
void mcv_en_debug_prints(
	int en
);

/**
 * @brief Globally enables/disables timing prints for libmodal-cv
 * 
 * @param en 1 to enable, 0 to disable
 */
void mcv_en_timing_prints(
	int en
);

/**
 * @brief Returns the system time in ns as type int64_t
 * 
 * @return The system time
 */
int64_t mcv_time_monotonic_ns(
	void
);

/**
 * @brief Util for converting (x,y) coords into a flattened index
 * 
 * @param x The x coordinate to convert to a flattened index 
 * @param y The y coordinate to convert to a flattened index 
 * @param width The width of the input image
 * @param height The height of the input image
 * @param row_major 1 if flat indicies are row major
 * @return The flattened index
 */
static inline int x_y_to_flat(
	int x,
	int y,
	int width,
	int height,
	int row_major
) {
	return row_major ? (y * width + x) : (x * height + y);
}

/**
 * @brief Util for converting a flattened index into (x,y) coords
 * 
 * @param flat_index 
 * @param width 
 * @param height 
 * @param row_major 
 * @param x 
 * @param y 
 */
static inline void flat_to_x_y(
	int flat_index,
	int width,
	int height,
	int row_major,
	int* x,
	int* y
) {
    if (row_major)
    {
        *x = flat_index % width;
        *y = flat_index / width;
    }
    else
    {
        *x = flat_index / height;
        *y = flat_index % height;
    }
}

#include <CL/cl.h>
/**
 * \brief Convert a 32-bit float to a 16-bit half float for use with OpenCL.
 *
 * @param f [in] - The 32-bit float to convert
 * @return the equivalent 16-bit half float
 */
cl_half float_to_half(float f);

/**
 * \brief Convert a 16-bit half float to a 32-bit float.
 *
 * @param f [in] - The 16-bit half float to convert
 * @return the equivalent 32-bit float
 */
cl_float half_to_float(cl_half f);

#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_COMMON_H
