/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @headerfile  dcm.h 
 *
 * @brief       Functions for computing Descriptor Calclulation and Matching 
 *              (DCM) using CVP hardware.
 * 
 *              See the voxl-test-dcm.cpp example for use case.
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2024
 */

#ifndef MODALCV_DCM_H
#define MODALCV_DCM_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#include "fpx.h"

#define MCV_CVP_DCM_MAX_DESCRIPTORS 1024

/**
 * @brief An object representing a generic handle for DCM in modalcv * using CVP 
 * hardware. 
 */
typedef void* mcv_cvp_dcm_handle;

/**
 * @brief Represents a 2D coordinate position on an image
 */
typedef struct mcv_cvp_dcm_pos_t
{
    uint16_t x; /**< Unsigned x coordinate */
    uint16_t y; /**< Unsigned y coordinate */
} __attribute__((packed)) mcv_cvp_dcm_pos_t;

/**
 * @brief Represents a DCM match
 */
typedef struct mcv_cvp_dcm_match_t
{
    uint16_t score; /**< from 0-63 */
    uint16_t index; /**< matching index from list */
} __attribute__((packed)) mcv_cvp_dcm_match_t;

/**
 * MAX SIZES
 */
#define MCV_CVP_MAX_POS_BUF_SIZE   (sizeof(mcv_cvp_dcm_pos_t)*MCV_CVP_DCM_MAX_DESCRIPTORS)
#define MCV_CVP_MAX_MATCH_BUF_SIZE (sizeof(mcv_cvp_dcm_match_t)*MCV_CVP_DCM_MAX_DESCRIPTORS)

/**
 * @brief Configuration info for the DCM process
 */
typedef struct mcv_cvp_dcm_config_t 
{
    int w; /**< The width of the input image */
    int h; /**< The height of the input image */
} mcv_cvp_dcm_config_t;

/**
 * @brief Output structure for DCM
 */
typedef struct mcv_cvp_dcm_output_t
{
    int n_descriptors; /**< Number of descriptors found */
    int n_matches; /**< Number of descriptor matches */
    void* descriptors; /**< The 256-bit descriptor data */
    mcv_cvp_dcm_match_t* matches; /**< Discovered DCM matches */
} mcv_cvp_dcm_output_t;

/**
 * @brief Creates an output object for DCM
 * 
 * @param handle The handle created my mcv_cvp_dcm_init
 * @param output Ptr to the output object
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_create_dcm_output(
    mcv_cvp_dcm_handle handle, 
    mcv_cvp_dcm_output_t* output
);

/**
 * @brief Destroys a created DCM output object
 * 
 * @param handle The handle created my mcv_cvp_dcm_init
 * @param output Ptr to the output object
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_destroy_dcm_output(
    mcv_cvp_dcm_handle handle, 
    mcv_cvp_dcm_output_t* output
);

/**
 * @brief Initializes the DCM process using the specified configuration
 * 
 * @param config A popuated confiuration object
 * @return The generated undistort handle if successful, NULL otherwise
 */
mcv_cvp_dcm_handle mcv_cvp_dcm_init(
    mcv_cvp_dcm_config_t config
);

/**
 * @brief De-initializes the DCM process
 * 
 * @param handle The handle from mcv_cvp_dcm_init
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_dcm_deinit(
    mcv_cvp_dcm_handle handle
);

/**
 * @brief Performs the DCM Process
 * 
 * @param handle The handle generated by mcv_cvp_dcm_init
 * @param img The handle to perform DCM on
 * @param fpx The 2D points to compute descriptors for
 * @param ref_descriptors The descriptors to attempt to match
 * @param n_ref_desc The number of reference descriptors
 * @param output The output object
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_dcm_process(
    mcv_cvp_dcm_handle handle,
    uint8_t* img,
    mcv_cvp_fpx_t* fpx,
    void* ref_descriptors,
    int n_ref_desc,
    mcv_cvp_dcm_output_t* output
);

#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_DCM_H
