/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

/**
 * @headerfile  dfs.h 
 *
 * @brief       Functions for computing depth from stereo (DFS)
 * 
 *              See the mcv-test-dfs.cpp example for use case.
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2024
 */

#ifndef MODALCV_DFS_H
#define MODALCV_DFS_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <modalcv/undistort_cvp.h>

/**
 * @brief An object representing a generic handle for DFS in modalcv
 * using CVP hardware. 
 */
typedef void* mcv_cvp_dfs_handle;

/**
 * @brief An object representing the undistortion + DFS pipeline
 * in modalcv using CVP hardware.
 */
typedef void* mcv_cvp_undistort_and_dfs_handle;

/**
 * @brief Configuration structure for performing DFS with CVP hardware
 * 
 * Note: Max Disparity must be multiple of 4 and is dependent on image width
 *    If the width <= 640, max disparity can be 64, 60, 56, ..., 4.
 *    If the width <= 1280, max disparity can be 32, 28, ..., 4.
 *    If the width <= 1920, max disparity can be 16, 12, ..., 4.
 */
typedef struct mcv_dfs_config_t{
	int width;             ///< width of the input image
	int height;            ///< height of the input image
	int n_disparity_levels;     ///< Maximum allowable disparity
	int starting_disparity_offset;  // Valid range is [0, 254-nMaxDisparity]
	int en_median_filter;
	int en_occlusion_filling;
	int occlusion_min_threshold;	///< [0-100] nOcclusionMinThreshold <= nOcclusionThreshold
	int occlusion_cost_threshold;	///< [0-100] nOcclusionMinThreshold <= nOcclusionThreshold
} mcv_dfs_config_t;

/**
 * @brief Initializes the DFS process using the specified configuration
 * 
 * @param config A populated undistort configuration object
 * @return The generated undistort handle if successful, NULL otherwise
 */
mcv_cvp_dfs_handle mcv_cvp_dfs_init(
	mcv_dfs_config_t config
);

/**
 * @brief De-initializes the DFS process 
 * 
 * @param handle The handle generated from mcv_undistort_init
 * @return 0 on success or -1 on failure
 */
int mcv_cvp_dfs_deinit(
	mcv_cvp_dfs_handle
);

/**
 * @brief Executes the DFS process
 * 
 * Note: disparity and occlusion masks are in a 4x4 raster, use other functions 
 * to convert to other desired output formats
 * 
 * @param left The left stereo image
 * @param right The right stereo image
 * @param disparity The output disparity mask
 * @param occlusion_mask The output occlusion mask
 * @return 0 on success or -1 on failure
 */
int mcv_cvp_dfs_process(
	mcv_cvp_dfs_handle handle,
	uint8_t* left,
	uint8_t* right,
	uint8_t* disparity,
	uint8_t* occlusion_mask
);

/**
 * @brief Initializes the Undistort+DFS pipeline using the specified config
 * 
 * @param undistort_config_l Configuration for the undistortion of L images
 * @param undistort_config_r Configuration for the undistortion of R images
 * @param dfs_config Configuration for the DFS process
 * @return A handle to be used for the process
 * 
 * @ingroup dfs_pipeline
 */
mcv_cvp_undistort_and_dfs_handle mcv_cvp_undistort_and_dfs_init(
	mcv_undistort_config_t undistort_config_l,
	mcv_undistort_config_t undistort_config_r,
	mcv_dfs_config_t dfs_config
);

/**
 * @brief Deinitializes the Undistort+DFS pipeline using the handle
 * 
 * @param undistort_and_dfs_handle The handle provided from the init fn
 * @return 0 if successful, -1 otherwise
 * 
 * @ingroup dfs_pipeline
 */
int mcv_cvp_undistort_and_dfs_deinit(
	mcv_cvp_undistort_and_dfs_handle undistort_and_dfs_handle
);

/**
 * @brief Executes the Undistort+DFS pipeline
 * 
 * @param handle The handle generated from the init function
 * @param left The left stereo image
 * @param right The right stereo image
 * @param disparity The output disparity values 
 * @param occlusion_mask The output occlusion mask
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_undistort_and_dfs_process(
	mcv_cvp_undistort_and_dfs_handle handle,
	uint8_t* left, 
	uint8_t* right, 
	uint8_t* disparity,
	uint8_t* occlusion_mask
);

/**
 * @brief Linearize the 4x4 raster output from DFS to a normal image
 * 
 * Note: This is slow, avoid using if possible
 * 
 * @param disp_in Input disparity image
 * @param w Image width
 * @param h Image height
 * @param disp_out Output linearlized disparity image
 * @return 0 if successful, -1 otherwise
 */
int mcv_dfs_linearize_disparity(
	uint8_t* disp_in,
	int w,
	int h,
	uint8_t* disp_out
);

/**
 * @brief Linearize the occlusion output from DFS to a normal image
 * 
 * @param occ_in Input occlusion image
 * @param w Image width
 * @param h Image hieght
 * @param occ_out Output linearized occlusion mask
 * @return 0 if successful, -1 otherwise
 */
int mcv_dfs_linearize_occlusion(
	uint8_t* occ_in,
	int w,
	int h,
	uint8_t* occ_out
);

#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_DFS_H
