/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/*
 * @headerfile  feature_tracker.h
 *
 * @brief       Header file for the modalcv feature tracker
 *
 *              Feature tracking in modalcv is a largely experimental feature. 
 *              In the context of ModalAI, this feature tracker serves 
 *              downstream VIO, and serves as an alternative to a more 
 *              traditional OpenCV-based tracker. This tracker makes use of 
 *              several hardware optimizations, namely outsourcing core tracking
 *              logic to the CVP. This can improve both processing times and CPU
 *              usage making the tracker very desiarable. But as previously 
 *              stated, the tracker is very experimental meaning there could be 
 *              regressions in performance compared to contemporary 
 *              alternatives.
 *
 * @author      Thomas Patton (thomas.patton@modalai.com) 
 * @date        2024 
 */

#ifndef MODALCV_FEATURE_TRACKER_H
#define MODALCV_FEATURE_TRACKER_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <modal_pipe.h>

#include "pyr_fpx.h"
#include "opt_flow.h"

#define MAX_N_FEATURES 9999

/**
 * @brief Generic handle for feature tracking in modalcv 
 *
 *        This void* handle holds proprietary information related to the CVP 
 *        in the context of feature tracking.
 */
typedef void* mcv_cvp_feature_tracker_handle;

/**
 * @brief Configuration for the feature tracking process 
 *        
 *        Since feature tracking is currently done through a combination of 
 *        pyramid feature poin extraction and optic flow, this configuration is 
 *        simply a nested configuration of those two configuration objects. This 
 *        struct should be populated before being passed for initialization.
 */
typedef struct mcv_cvp_feature_tracker_config_t {

    /** @brief The configuration for pyramid feature point extraction */
    mcv_cvp_pyr_fpx_config_t pyr_fpx_config;

    /** @brief The configuration for optic flow */
    mcv_cvp_opt_flow_config_t opt_flow_config;

    /** @brief Which octave on the pyramid to perform optic flow on */
    int opt_flow_octave;

    /** @brief Indicates whether pyr_fpx and opt_flow images are different */
    bool two_img_input;

    /** @brief Binary mask of size (wxh) corresponding to where features are 
     * valid. If NULL, mask is assumed to be all 1's */
    uint8_t* feature_mask;

} mcv_cvp_feature_tracker_config_t;

/**
 * @brief Structure representing the output of the feature tracker.
 *
 *        This structure is made to closely mirror the `vft_feature` struct in
 *        the voxl-feature-tracker repository for easy data transfer.
 */
typedef struct mcv_cvp_feature_tracker_output_t {
    /** @brief Unique id for the feature point. The same feature should have the
     *        same id on consecutive frames. */
    int64_t id;

    /** @brief Sub-pixel refined x-coordinate of the feature */
    float x;

    /** @brief Sub-pixel refined y-coordinate of the feature */
    float y;

    /** @brief Previous x coordinate for the feaure */
    float x_prev;

    /** @brief Previous y coordinate for the feaure */
    float y_prev;

    /** @brief The 256-bit binary descriptor for the feature */
    unsigned char descriptor[32];

    /** @brief The feature's age in number of frames */
    int32_t age;

    /** @brief The generated score between [0-100] for the feature */
    int8_t score;

    /** @brief A bitmask for present pyramid levels */
    int8_t pyr_lvl_mask;

    int32_t  cvp_scores[3];
    int8_t   reserved_1;
    uint32_t reserved_2;
} __attribute__((packed)) mcv_cvp_feature_tracker_output_t;

/**
 * @brief Initialize the feature tracker with a given config
 * 
 * @param config A populated configuration object
 * @return mcv_cvp_feature_tracker_handle 
 */
mcv_cvp_feature_tracker_handle
mcv_cvp_feature_tracker_init(mcv_cvp_feature_tracker_config_t config);

/**
 * @brief De-initialize the feature tracker 
 * 
 * @param handle The handle obtained from the initialization function
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_feature_tracker_deinit(mcv_cvp_feature_tracker_handle handle);

/**
 * @brief Finds and tracks features from image to image
 * 
 * @param handle The image generated by the init function
 * @param img The current image
 * @param output_size The size of the output
 * @param output The output structure
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_feature_tracker_process(
	int cid,
    mcv_cvp_feature_tracker_handle   handle,
    uint8_t*                         image1, 
    uint8_t*                         image2,
    int16_t                          n_feats_to_select,
    void*                            prev_selected_ids,
    void*                            feature_distribution,
    int16_t*                         output_size,
    mcv_cvp_feature_tracker_output_t output[MAX_N_FEATURES]);


void cvp_set_pixel_search_radius(double val);
void cvp_set_octave_search_radius(double val);
void cvp_set_en_id_sort(bool val);
void cvp_set_en_cam_display(int val);

typedef enum mcv_cvp_motion_vec_interp_t {
    BILINEAR = 0,
    BICUBIC
} mcv_cvp_motion_vec_interp_t;

typedef enum mcv_cvp_var_state_t{
    PREV = 0,
    CURRENT = 1,
    N_VAR_STATES
} mcv_cvp_var_state_t;


#ifdef __cplusplus
}
#endif

#endif // MODALCV_FEATURE_TRACKER_H
