/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @headerfile  fpx.h 
 *
 * @brief       Functions for computing Feature Point Extraction (FPX) using  
 *              CVP hardware.
 * 
 *              See the voxl-test-fpx.cpp example for use case.
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2024
 */

#ifndef MODALCV_FPX_H
#define MODALCV_FPX_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#define MAX_FPX_CAMS 4

/**
 * @brief An object representing a generic handle for feature point extraction
 * in modalcv using CVP hardware
 */
typedef void* mcv_cvp_fpx_handle;

/**
 * @brief Different modes for extracting feature points
 */
typedef enum mcv_cvp_fpx_mode_t {
    MCV_CVP_FPX_PEAK_8x8 = 0,	/**< Extracts the top feature point from 8x8 */
    MCV_CVP_FPX_ZONE /**< Top 8 features per zone */
} mcv_cvp_fpx_mode_t;

/**
 * @brief Different modes for non-maximum suppression
 */
typedef enum mcv_cvp_fpx_nms_t {
    MCV_CVP_FPX_3_TAP_NMS = 0, /**< three-tap (3×3 filter) NMS */
    MCV_CVP_FPX_5_TAP_NMS /**< five-tap (5×5 filter) NMS */
} mcv_cvp_fpx_nms_t;

/**
 * @brief Configuration for performing feature point extraction
 */
typedef struct mcv_cvp_fpx_config_t{
    int width; /**< width of the input image */
    int height; /**< height of the input image */
    mcv_cvp_fpx_mode_t mode; /**< fpx mode */
    mcv_cvp_fpx_nms_t nms_mode; /**< nms mode */
    uint32_t score_threshold; /**< threshold for FPX score */
    uint8_t robustness; /**< lower robustness setting produces more points */
} mcv_cvp_fpx_config_t;

/**
 * @brief Represents a single extracted feature
 */
typedef struct mcv_cvp_fpx_feature_t{
    uint16_t x; /**< x coordinate of the feature */
    uint16_t y; /**< y coordinate of the feature */
    uint16_t score; /**< feature score */
    uint16_t reserved;
} __attribute__((packed)) mcv_cvp_fpx_feature_t;

/**
 * @brief Object to store the output of feature point extraction
 */
typedef struct mcv_cvp_fpx_t {
    uint32_t n_points; /**< number of extracted points */
    uint32_t max_score; /**< max extraction score */        
    mcv_cvp_fpx_feature_t* features; /**< extracted feature points */                       
} mcv_cvp_fpx_t;

/**
 * @brief Initializes the FPX process using the specified configuration
 * 
 * Internally, this initialization involves configuring and allocating 
 * several hardware buffers
 * 
 * @param config The populated configuration object
 * @return The generated FPX handle if successful, NULL otherwise
 */
mcv_cvp_fpx_handle mcv_cvp_fpx_init(
    mcv_cvp_fpx_config_t config
);

/**
 * @brief Deinializes the FPX process
 * 
 * @param handle The handle returned from mcv_cvp_fpx_init
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_fpx_deinit(
    mcv_cvp_fpx_handle handle
);

/**
 * @brief Runs the FPX process, writing output data to an object
 * 
 * @param handle The handle returned from mcv_cvp_fpx_init
 * @param img The input image to compute FPX on
 * @param output FPX object to store the output
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_fpx_process(
    mcv_cvp_fpx_handle handle, 
    uint8_t* img, 
    mcv_cvp_fpx_t* output
);

/**
 * @brief Changes the zone dimensions for the FPX process
 * 
 * @param handle The handle returned from mcv_cvp_fpx_init
 * @param x Frame width
 * @param y Frame height
 * @return 0 if successful, -1 for generic error, -2 if not in zone mode, -3 
 * if invalid size
 */
int mcv_cvp_fpx_change_zone_dims(
    mcv_cvp_fpx_handle handle,
    uint32_t x,
    uint32_t y
);

/**
 * @brief Creates an output FPX object
 * 
 * @param handle The handle returned from the initialization method
 * @param fpx Output FPX object
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_create_fpx(mcv_cvp_fpx_handle handle, mcv_cvp_fpx_t* fpx);

/**
 * @brief Destroys an output FPX object
 * 
 * @param handle The handle returned from the initialization method
 * @param fpx Output FPX object to be destroyed
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_destroy_fpx(mcv_cvp_fpx_handle handle, mcv_cvp_fpx_t* fpx);

#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_FPX_H
