/*******************************************************************************
 * Copyright 2023 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/
			
/**
 * @headerfile  lens.h
 *
 * @brief       Helper functions for working with different types of lenses 
 *
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2023 
 */

#ifndef MODALCV_LENS_H
#define MODALCV_LENS_H


#ifdef __cplusplus
extern "C" {
#endif

#include <modalcv/undistort_cvp.h>


/**
 * @brief Generic structure for a mono lens
 * 
 */
typedef struct mcv_lens_params_t {
	int w;
	int h;
	float fx;              //< Focal length (x) in pixels    
	float fy;              //< Focal length (y) in pixels
	float cx;              //< Optical center (x) in pixels    
	float cy;              //< Optical center (y) in pixels
	int n_coeffs;          //< 4 for fisheye, 5 for OpenCV pinhole      
	float D[12];           //< Distortion coefficients
	int is_fisheye;        //< 1 for 4-param fisheye

    mcv_cvp_undistort_grid_t undistort_grid; //< Output undistortion grid
} mcv_lens_params_t;

/**
 * @brief Generic structure for a stereo lens
 * 
 */
typedef struct mcv_stereo_params_t {
	int w;               //< Width of the image     
	int h;               //< Height of the image     
	float fxrect;        //< Rectified focal length (x)            
	float fyrect;        //< Rectified focal length (y)      
	float cxrect;        //< Rectified optical center (x)     
	float cyrect;        //< Rectified optical center (y)    
	float baseline_m;    //< Distance between stereo cameras        

    mcv_cvp_undistort_grid_t undistort_grid_r; //< Output undistortion grid
    mcv_cvp_undistort_grid_t undistort_grid_l; //< Output undistortion grid
} mcv_stereo_params_t;

/**
 * @brief Creates a single undistort grid from an intrinsics file
 * 
 * This is intended to be used for mono cameras where only an intrinsics
 * file is needed. For two stereo cameras, the function below should be used
 * 
 * @param intrinsics_file Intrinsics file to be read
 * @param zoom Typically 0.6 for fisheye and 0.9 for other lenses
 * @param lens_params Output object for lens parameter storage
 * @param undistort_grid Output undistortion grid
 * 
 */
int mcv_create_lens_undistort_grid(const char* intrinsics_file, float zoom,
	mcv_lens_params_t* lens_params, mcv_cvp_undistort_grid_t* undistort_grid);

/**
 * @brief Creates left/right undistort grids from in/extrinsics files
 * 
 * @param intrinsics_file Intrinsics file to be read
 * @param extrinsics_file Extrinsics file to be read
 * @param stereo_params Output object for stereo params
 * @param undistort_grid_left Output undistortion grid for left camera
 * @param undistort_grid_right Output undistortion grid for right camera
 * @return int 
 */
int mcv_create_stereo_undistort_grid(const char* intrinsics_file, 
	const char* extrinsics_file, mcv_stereo_params_t* stereo_params, 
	mcv_cvp_undistort_grid_t* undistort_grid_left, 
	mcv_cvp_undistort_grid_t* undistort_grid_right);
	
/**
 * @brief Populates a lens parameter object based on an intrinsics file
 * 
 * Essentially reads the file line by line and finds variables corresponding
 * to lens features
 * 
 * @param intrinsics_file The intrinsics file to be read
 * @param lens_params Output lens parameter object
 * @return 0 if successful, -1 otherwise
 */
int init_lens_params(const char* intrinsics_file, 
	mcv_lens_params_t* lens_params);

/** UNUSED **/
// int mcv_pull_mono_intrinsics(const char *file, mcv_lens_params_t *ret);
// int mcv_pull_stereo_intrinsics(const char *file, mcv_lens_params_t *ret1, mcv_lens_params_t *ret2);

// // make sure you allocate 9 floats for the rotation and 3 for translation
// int mcv_pull_extrinsics(const char *file, float* rotation, float* translation);

#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_LENS_H