/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/
/*
 * @headerfile  ncc.h
 *
 * @brief       Header files for computing normalized cross-correlation (NCC) on 
 *              CVP hardware.
 *
 * @author      Thomas Patton (thomas.patton@modalai.com) 
 * @date        2023 
 */


#ifndef MODALCV_NCC_H
#define MODALCV_NCC_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#include <modal_pipe.h>

#include "fpx.h"


#define MCV_CVP_NCC_MAX_TEMPLATES         750
#define MCV_CVP_NCC_PATCH_SIZE            64

/**
 * @brief Represents a generic handle for performing NCC
 */
typedef void* mcv_cvp_ncc_handle;

/**
 * @brief Whether to perform NCC in patch-based or frame-based mode
 * 
 * @param MCV_CVP_NCC_PATCH_MODE Perform NCC in patch-based mode. In this mode
 * the user must assemble their own templates for NCC search
 * 
 * @param MCV_CVP_NCC_FRAME_MODE Perform NCC in frame-based mode. In this mode
 * the previous frame is used as a template for NCC search
 */
typedef enum mcv_cvp_ncc_mode 
{
    MCV_CVP_NCC_PATCH_MODE = 0, 
    MCV_CVP_NCC_FRAME_MODE      
} mcv_cvp_ncc_mode;

/**
 * @brief Whether to return all scores or only the best score
 * 
 * @param MCV_CVP_NCC_BEST_SCORE Return only the best score from NCC
 * @param MCV_CVP_NCC_ALL_SCORE Returns all scores from NCC. Output must 
 * then be linearized 
 */
typedef enum mcv_cvp_ncc_score_type 
{
    MCV_CVP_NCC_BEST_SCORE = 0, /**< Gives output for only the best score */
    MCV_CVP_NCC_ALL_SCORE       /**< Gives all scores */
} mcv_cvp_ncc_score_type;

/**
 * @brief Object representing a single search window for NCC
 * 
 * @param ref_loc_x Top left of 18×18 window position x in the reference image.
 * @param ref_loc_y Top left of 18×18 window position y in the reference image.
 * @param src_loc_x (Frame mode only) Top left of current source 8×8 window 
 * position x.
 * @param src_loc_y (Frame mode only) Top left of current source 8×8 window 
 * position y. 
 */
typedef struct mcv_cvp_ncc_search_window 
{
    uint64_t ref_loc_x;     /**< Top left of 18×18 window position x in the reference image */
    uint64_t ref_loc_y;     /**< Top left of 18×18 window position y in the reference image */
    uint64_t src_loc_x;     /**< Top left of current source 8×8 window position x.For FRAME_MODE only, otherwise ignored */
    uint64_t src_loc_y;     /**< Top left of current source 8×8 window position y.For FRAME_MODE only, otherwise ignored */
} mcv_cvp_ncc_search_window;

/**
 * @brief Object representing a single result window for NCC
 * 
 * @param match_loc_x Top left of 18×18 window position x in the reference image.
 * @param match_loc_y Top left of 18×18 window position y in the reference image.
 * @param ref_loc_x Top left of current source 8×8 window position x. For FRAME_MODE only, in PATCH_MODE, this parameter is copied from the search window src_loc_x
 * @param ref_loc_y Top left of current source 8×8 window position y. For FRAME_MODE only, in PATCH_MODE, this parameter is copied from the search window src_loc_y
 * @param robustness Robustness value of NCC.
 * @param subpel Indicates whether subpel was enabled (from configuration).
 * 
 * @ingroup mcv_cvp_types
 */
typedef struct mcv_cvp_ncc_result_window {
    uint64_t match_loc_x;   
    uint64_t match_loc_y;   
    uint64_t src_loc_x;     
    uint64_t src_loc_y;     
    uint64_t robustness;    
    uint64_t subpel;        
} mcv_cvp_ncc_result_window;


/**
 * @brief Configuration object for NCC process
 * 
 * @param width Width of the input image
 * @param height Height of the input image
 * @param ncc_mode An mcv_cvp_ncc_mode as defined above
 * @param score_type An mcv_cvp_ncc_score_type as defined above
 * @param en_robustness Boolean to enable robustness
 * @param en_subpel Boolean to enable subpel
 * 
 * @ingroup mcv_cvp_types
 */
typedef struct mcv_cvp_ncc_config_t 
{
    int width;
    int height;
    mcv_cvp_ncc_mode ncc_mode;
    mcv_cvp_ncc_score_type score_type;
    int en_robustness;                 // Specifies whether to enable robustness. By default, robustness is disabled (FALSE).
    int en_subpel;                     // Specifies whether to enable subpel. By default, subpel is enabled (TRUE).

    int input_ubwc = 0;                // if the input ref image is in UBWC format. Recommended to leave at 0
} mcv_cvp_ncc_config_t;

/**
 * @brief 
 * 
 * n_results is the integer number of features that were attempted to be 
 * matched. this is passed directly from the input fpx structure's n_points.
 * 
 * valid_match is a binary mask representing if there
 * 
 */
typedef struct mcv_cvp_ncc_output_t {
    int n_results;   
    int valid_match[MCV_CVP_NCC_MAX_TEMPLATES];
    mcv_cvp_ncc_result_window* results;
    int16_t* scores;
} mcv_cvp_ncc_output_t;


// init
mcv_cvp_ncc_handle mcv_cvp_ncc_init(
    mcv_cvp_ncc_config_t config
);

// deinit
int mcv_cvp_ncc_deinit(
    mcv_cvp_ncc_handle handle
);

// process
int mcv_cvp_ncc_process(
    mcv_cvp_ncc_handle handle,
    uint8_t* src_img,
    uint8_t* ref_img,
    mcv_cvp_fpx_t* fpx,
    mcv_cvp_ncc_output_t* output
);

// create output structure
int mcv_cvp_create_ncc_output(
    mcv_cvp_ncc_handle handle,
    mcv_cvp_ncc_output_t* output
);

// tear down above
int mcv_cvp_destroy_output(
    mcv_cvp_ncc_output_t* output
);

/**
 *
 */
int mcv_cvp_ncc_linearize_scores(
    int16_t* scores_in,
    int n_requests,
    int16_t* scores_out
);

#ifdef __cplusplus
}
#endif

#endif // MODALCV_NCC_H
