/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @headerfile  opt_flow.h 
 *
 * @brief       Functions for computing Semi-Dense Optical Flow using CVP 
 *              hardware.
 * 
 *              See the mcv-test-opt-flow.cpp example for use case.
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2024
 */

#ifndef MODALCV_OPT_FLOW_H
#define MODALCV_OPT_FLOW_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

/**
 * @brief An object representing a generic handle for optical flow
 * in modalcv using CVP hardware
 */
typedef void* mcv_cvp_opt_flow_handle;

/**
 * @brief Configuration info for optic flow
 *
 * Unless you're sure of what you're doing, leave input_ubwc as 0
 */
typedef struct mcv_cvp_opt_flow_config_t 
{
    int width; /**< width of the input image */
    int height; /**< height of the input image */ 
    int en_stats; /**< bool, enable/disable statistics */ 
    int input_ubwc = 0; /**< recommended to leave at 0 */
} mcv_cvp_opt_flow_config_t;

/**
 * @brief Represents a generic motion vector
 * 
 * This struct is designed to work nicely with the CVP
 * conf ranges from -8 to +8 but is essentially useless, ignore it
 */
typedef struct mcv_cvp_motion_vector_t
{
    int32_t mv_x      : 9; /**< horizontal mv for 8x8 block */
    int32_t mv_y      : 7; /**< vertical mv for 8x8 block */    
    int32_t reserved  : 12; /**< reserved */  
    int32_t conf      : 4; /**< motion vector confidence */ 
} mcv_cvp_motion_vector_t;

/**
 * @brief Represents the statistics for an 8x8 block of optic flow
 */
typedef struct mcv_cvp_opt_flow_stats_t
{
    uint16_t variance;    // variance of the 8x8 block
    uint8_t  mean;        // pixel intensity mean of the 8x8 block
    uint8_t  reserved;   
    uint16_t best_sad;    // best sum-of-absolute-difference
    uint16_t n_sad;       // sum-of-absolute-difference if there is no motion vec
} mcv_cvp_opt_flow_stats_t;

/**
 * @brief Generic output object for optical flow
 */
typedef struct mcv_cvp_opt_flow_output_t
{
    mcv_cvp_motion_vector_t* mvs; /**< motion vectors */
    mcv_cvp_opt_flow_stats_t* stats; /**< stats */
} mcv_cvp_opt_flow_output_t;

/**
 * @brief Creates an optical flow output object
 *
 * @param handle The optical flow handle
 * @param output Ptr to object for population
 */
int mcv_cvp_create_opt_flow_output(
    mcv_cvp_opt_flow_handle handle,
    mcv_cvp_opt_flow_output_t* output
);

/**
 * @brief Destroys an optical flow output object
 *
 * @param handle The optical flow handle
 * @param output Ptr to object to destroy
 */
int mcv_cvp_destroy_opt_flow_output(
    mcv_cvp_opt_flow_handle handle,
    mcv_cvp_opt_flow_output_t* output
);


/**
 * @brief Uses a configuration object to generate a handle for optic flow
 * 
 * @param config A populated optic flow configuration object
 * @return The handle to use for optic flow
 */
mcv_cvp_opt_flow_handle mcv_cvp_opt_flow_init(
    mcv_cvp_opt_flow_config_t config
);

/**
 * @brief De-initializes the optic flow process
 * 
 * @param handle The optic flow handle provided from the init function
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_opt_flow_deinit(
    mcv_cvp_opt_flow_handle handle
);

/**
 * @brief Performs the optic flow process between two images
 * 
 * Returns the motion vector and statistics data from the current image
 * to the previous image. Stores `curr_img` into memory so it can be used as
 * the previous img on next iter
 * 
 * @param handle The optic flow handle from the above function
 * @param prev_img Image data for the "previous" frame
 * @param curr_img Image data for the "current" frame
 * @param output Output object to populate
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_opt_flow_process_from_img(
    mcv_cvp_opt_flow_handle handle,
    uint8_t* prev_img, 
    uint8_t* curr_img, 
    mcv_cvp_opt_flow_output_t* output
);

/**
 * @brief Performs the optic flow process between an img and the image stored in 
 * memory.
 * 
 * Returns the motion vector and statistics data from the current image
 * to the previous image. Stores `curr_img` into memory so it can be used as
 * the previous img on next iter
 * 
 * NOTE: You should use the above function at least once to actually register
 * an image into memory. Otherwise this function will not work accurately.
 * 
 * @param handle The optic flow handle from the above function
 * @param curr_img Image data for the "current" frame
 * @param output Output object to populate
 * @return 0 if successful, -1 otherwise
 */
int mcv_cvp_opt_flow_process_from_mem(
    mcv_cvp_opt_flow_handle handle,
    uint8_t* curr_img, 
    mcv_cvp_opt_flow_output_t* output
);


#ifdef __cplusplus
}
#endif

#endif // MODALCV_OPT_FLOW_H
