/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @headerfile  pyr.h <modalcv/pyr.h>
 *
 * @brief       Functions for computing image pyramiding in modalcv using CVP
 *              hardware.
 *  
 *              See the voxl-test-pyr.cpp example for use case.
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2023
 */


#ifndef MODALCV_PYR_H
#define MODALCV_PYR_H

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif


#define MCV_CVP_MAX_PYRAMID_OCTAVES 5
#define MCV_CVP_MAX_PYRAMID_SCALES_PER_OCTAVE 4
#define MCV_CVP_MAX_PYRAMID_LEVELS (MCV_CVP_MAX_PYRAMID_OCTAVES * MCV_CVP_MAX_PYRAMID_SCALES_PER_OCTAVE)

/**
 * @brief An object representing a generic handle for pyramid downscaling in
 * modalcv using CVP hardware
 */
typedef void* mcv_cvp_pyr_handle;

/**
 * @brief Configuration object for pyramid downscaling
 */
typedef struct mcv_cvp_pyr_config_t {
	int width; /**< input width */
	int height; /**< input height */
	uint32_t n_octaves; /**< number of octaves */
	uint32_t n_scales_per_octave; /**< number of scales for each octave */
	int output_ubwc; /**< output ubwc format*/
} mcv_cvp_pyr_config_t;

/**
 * @brief Generic image pyramid
 */
typedef struct mcv_cvp_pyr_t {
	int width[MCV_CVP_MAX_PYRAMID_LEVELS]; /**< width of image per level */	
	int height[MCV_CVP_MAX_PYRAMID_LEVELS]; /**< height of the image per level */ 
	uint8_t* image_pyramids[MCV_CVP_MAX_PYRAMID_LEVELS]; /**< array of image pyr */
} mcv_cvp_pyr_t;

mcv_cvp_pyr_handle mcv_cvp_pyr_init(
    mcv_cvp_pyr_config_t config
);




int mcv_cvp_pyr_deinit(mcv_cvp_pyr_handle handle);
int mcv_cvp_create_pyr(mcv_cvp_pyr_handle handle, mcv_cvp_pyr_t* pyr);
int mcv_cvp_destroy_pyr(mcv_cvp_pyr_handle handle, mcv_cvp_pyr_t* pyr);
int mcv_cvp_pyr_process(mcv_cvp_pyr_handle handle, uint8_t* img, 
    mcv_cvp_pyr_t* pyr);

#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_PYR_H
