/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @headerfile  pyr_fpx.h <modalcv/pyr_fpx.h>
 *
 * @brief       Functions for computing Pyramid Feature Point Extraction (FPX) 
 *              using CVP hardware.
 * 
 *              See the voxl-test-pyr-fpx.cpp example for use case.
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2023
 */

#ifndef MODALCV_PYR_FPX_H
#define MODALCV_PYR_FPX_H

#include <stdint.h>
#include <modalcv/fpx.h>
#include <modalcv/pyr.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief An object representing a generic handle for pyramid feature point
 * extraction in modalcv using CVP hardware
 *
 * @ingroup mcv_cvp_types
 */
typedef void* mcv_cvp_pyr_fpx_handle;

/**
 * @brief Configuration object for pyramid feature point extraction
 *
 * @ingroup mcv_cvp_types
 */
typedef struct mcv_cvp_pyr_fpx_config_t {
	int width;						// input width
	int height;						// input height
	uint32_t n_octaves;				// number of octaves. Image size is halved between adjacent octaves
	uint32_t n_scales_per_octave;	// number of scales per octave, including the octave. Number of levels is equal to nScalesPerOctave * nOctaves.
	mcv_cvp_fpx_mode_t mode;		// MCV_CVP_FPX_PEAK_8x8 or MCV_CVP_FPX_ZONE
	mcv_cvp_fpx_nms_t nms_mode;		// MCV_CVP_FPX_3_TAP_NMS or MCV_CVP_FPX_5_TAP_NMS (default)
	uint32_t score_threshold;		//
	uint8_t robustness;             // 0-127, default 10
	uint8_t score_shift;            // 0, 4, 8, or 12
	int enable_fpx[MCV_CVP_MAX_PYRAMID_OCTAVES]; // flag of whether or not to extract at this level
	int zone_dims[MCV_CVP_MAX_PYRAMID_OCTAVES];  // zone dim for each octave, only applicable in zone mode
	int output_ubwc;             // < LEAVE THIS AT ZERO
} mcv_cvp_pyr_fpx_config_t;


/**
 * get the default set of config values for the pyr_fpx module.
 * note you should definitely make sure to set the width and height correctly
 * as every camera is different!!
 */
mcv_cvp_pyr_fpx_config_t mcv_cvp_pyr_fpx_default_config(void);


/**
 * @brief Generic feature point extraction pyramid
 *
 * Note that for this structure, the item at index 0 for each member corresponds
 * to the original image. So n_points[0] will have the number of extracted
 * points for the original image, n_points[1] will have the number of extracted
 * points for the first pyramid level, etc...
 *
 * @ingroup mcv_cvp_types
 */
typedef struct mcv_cvp_pyr_fpx_t {
	int n_points[MCV_CVP_MAX_PYRAMID_LEVELS];
	int width[MCV_CVP_MAX_PYRAMID_LEVELS];					    // width of image per level
	int height[MCV_CVP_MAX_PYRAMID_LEVELS];					    // height of image per level
	uint8_t* image_pyramids[MCV_CVP_MAX_PYRAMID_LEVELS];	    // array of image pyramids, malloced by mcv_cvp_pyr_init
	mcv_cvp_fpx_feature_t* fpx[MCV_CVP_MAX_PYRAMID_LEVELS];     // feature points per level
} mcv_cvp_pyr_fpx_t;

/**
 * @brief Initializes the Pyramid FPX process using the specified configuration
 *
 * @param config A populated configuration object
 * @return The generated handle if successful, NULL otherwise
 *
 * @ingroup pyr_fpx
 */
mcv_cvp_pyr_fpx_handle mcv_cvp_pyr_fpx_init(mcv_cvp_pyr_fpx_config_t config);

/**
 * @brief Deinitializes the Pyramid FPX process
 *
 * @param handle The handle returned from mcv_cvp_pyr_fpx_init
 * @return 0 if successful, -1 otherwise
 *
 * @ingroup pyr_fpx
 */
int mcv_cvp_pyr_fpx_deinit(mcv_cvp_pyr_fpx_handle handle);

/**
 * @brief Runs the Pyramid FPX process
 *
 * Here the user should pass in a blank mcv_cvp_pyr_fpx object. During the
 * process, this object will be populated with the appropriate values and
 * pointers.
 *
 * @param handle The handle returned from mcv_cvp_pyr_fpx_init
 * @param img The input image to perform Pyramid FPX on
 * @param pyr_fpx An empty output structure
 * @return 0 if successful, -1 otherwise
 *
 * @ingroup pyr_fpx
 */
int mcv_cvp_pyr_fpx_process(mcv_cvp_pyr_fpx_handle handle, uint8_t* img,
	mcv_cvp_pyr_fpx_t* pyr_fpx);


/**
 * @brief Utility function to create a feature extraction pyramid object
 * 
 * @param handle The handle returned from mcv_cvp_pyr_fpx_init
 * @param pyr_fpx A pointer to the output pyr_fpx object
 * @return 0 if successful, -1 otherwise
 * 
 * @ingroup pyr_fpx_util
 */
int mcv_cvp_create_pyr_fpx(mcv_cvp_pyr_fpx_handle handle, mcv_cvp_pyr_fpx_t* pyr_fpx);

/**
 * @brief Utility function to destroy a feature extraction pyramid object
 * 
 * @param handle The handle returned from mcv_cvp_pyr_fpx_init
 * @param pyr_fpx A pointer to the output pyr_fpx object
 * @return 0 if successful, -1 otherwise
 * 
 * @ingroup pyr_fpx_util
 */
int mcv_cvp_destroy_pyr_fpx(mcv_cvp_pyr_fpx_handle handle, mcv_cvp_pyr_fpx_t* pyr_fpx);

#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_PYR_FPX_H
