/*******************************************************************************
 * Copyright 2023 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

/*
 * @headerfile  timer.h
 *
 * @brief       Timer utility for libmodal-cv or otherwise
 * 
 * Generally speaking, this module can be thought of like a traditional 
 * stopwatch. `mcv_timer_reset` will clear the timer to zero. Whenever you want
 * to hit the "split" button on your stopwatch, call `mcv_timer_log` with a 
 * message about the thing you just timed. And then at any point, call 
 * `mcv_timer_summary` to get a nice printout of the things you wished to time.
 * 
 * Example:
 * 
 * int main() {
 * 
 *     // create and reset the timer
 *     mcv_timer_t timer;
 *     mcv_timer_reset();
 * 
 *     mcv_timer_log(&timer, "start");
 * 
 *     // call and time function "a"
 *     a();
 *     mcv_timer_log(&timer, "a");
 * 
 *     // call and time function "b"
 *     b();
 *     mcv_timer_log(&timer, "b");
 * 
 *     mcv_timer_summary()
 *     // ... timer printout to stdout
 * }
 * 
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com) 
 * @date        2023 
 */

#ifndef MODALCV_TIMER_H
#define MODALCV_TIMER_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include <modalcv/common.h>

#define MAX_LOGS   100                // maximum number of splits
#define MSG_WIDTH  30                 // max width for a text message
#define TIME_WIDTH 7                  // max width for a timestamp

/**
 * @brief A timer to record int64_t timestamps
 */
typedef struct mcv_timer_t {
    int64_t timestamps[MAX_LOGS];    // the timestamps
    char*   messages[MAX_LOGS];      // messages associated with the timestamps
    int     n_timestamps;            // number of current timestamps
} mcv_timer_t;

/**
 * @brief Clears the timestamp register 
 * 
 * @param timer 
 */
static void mcv_timer_reset(
    mcv_timer_t* timer
) {
    timer->n_timestamps = 0;
    for(int i = 0; i < MAX_LOGS; i++) {
        timer->timestamps[i] = 0;
        timer->messages[i]   = NULL;
    }
}

static void mcv_timer_log(
    mcv_timer_t* timer,
    const char* message
) {
    int i = timer->n_timestamps;
    timer->timestamps[i] = (double)mcv_time_monotonic_ns();
    timer->messages[i]   = strdup(message);
    timer->n_timestamps++;
}


static void mcv_timer_summary(
    mcv_timer_t* timer
) { 
    double total_time = (timer->timestamps[timer->n_timestamps-1] - timer->timestamps[0]);
    printf("+-%-*s-+-%-*s-+-%-*s-+\n", MSG_WIDTH, "------------------------------", 
        TIME_WIDTH, "----------", 9, "---------");
    printf("| \033[1m%-30s\033[0m | \033[1m%-10s\033[0m | \033[1m%-9s\033[0m |\n", "Description", 
        "Time", "Percent");
    printf("+-%-*s-+-%-*s-+-%-*s-+\n", MSG_WIDTH, "------------------------------", 
        TIME_WIDTH, "----------", 9, "---------");

    for (int i = 1; i < timer->n_timestamps; i++) {
        double elapsed = (double)(timer->timestamps[i] - timer->timestamps[i-1]);
        double percent = (elapsed / total_time) * 100;
        printf("| %-30s | %8.3fms | %8.2f%% |\n", timer->messages[i], elapsed / 1000000, percent);
    }
    printf("+-%-*s-+-%-*s-+-%-*s-+\n", MSG_WIDTH, "------------------------------", 
        TIME_WIDTH, "----------", 9, "---------");
    printf("| \033[1m%-30s\033[0m | \033[1m%8.3fms\033[0m |\n", "Total", total_time / 1000000);
    printf("+-%-*s-+-%-*s-+\n", MSG_WIDTH, "------------------------------", 
        TIME_WIDTH, "----------");
}


#ifdef __cplusplus
}
#endif

#endif // MODALCV_TIMER_H