/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

/**
 * @headerfile  undistort_cpu.h <libmodal-cv/undistort_cpu.h>
 *
 * @brief       Functions for undistorting and rectifying images
 *              using the CPU hardware.
 * 
 *              See the voxl-test-undistort.cpp example for use case.
 * 
 * @author      James Strawson
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2023
 */


#ifndef MODALCV_UNDISTORT_CPU_H
#define MODALCV_UNDISTORT_CPU_H


#ifdef __cplusplus
extern "C" {
#endif


#include <stdint.h>
#include <pthread.h>
#include "lens.h"

////////////////////////////////////////////////////////////////////////////////
// Bilinear interpolation undistort on the CPU
////////////////////////////////////////////////////////////////////////////////



// undistort_map_t points to a big array of these, 1 per pixel
typedef struct mcv_bilinear_lookup_t{
	int16_t		I[2]; // backwards map to top left corner of 2x2 square
	uint8_t		F[4]; // 4 coefficients for 4 corners of square
} mcv_bilinear_lookup_t;


typedef struct mcv_undistort_map_t{
	int initialized;
	int w;			// image width
	int h;			// image height
	float fxrect;	// rectified focal length
	float fyrect;	// rectified focal length
	float cxrect;	// rectified principle point (center) x
	float cyrect;	// rectified principle point (center) y
	mcv_bilinear_lookup_t* L; // pointer to lookup table
} mcv_undistort_map_t;

#define MCV_UNDISTORT_MAP_INITIALIZER {\
	.initialized = 0,\
	.w = 0,\
	.h = 0,\
	.fxrect = 0,\
	.fyrect = 0,\
	.cxrect = 0,\
	.cyrect = 0,\
	.L = NULL\
}


/**
 * initialize an undistort map from a lens cal model zoom will change the focal
 * length and is useful to reduce loss of data on extremely distorted cameras
 * like fisheye lenses. For a fisheye lens we recommend a zoom of 0.6 which
 * strikes a nice balance of trimming off the corners of the image and leaving
 * just a little black area near the edges. Leaving this at 1 keeps the focal
 * length the same but would result in lost data as the image is "stretched" out
 * during undistortion. 0.9 works well for most lenses to prevent data loss near
 * the corners.
 *
 * @param[out] map   The final computed map
 * @param[in]  lens  The lens params
 * @param[in]  zoom  typically 0.6 for fisheye, 0.9 for other lenses
 *
 * @return     0 on success, -1 on failure
 */
int mcv_init_undistort_map_lens(mcv_undistort_map_t* map, mcv_lens_params_t lens, float zoom);

// initialize an undistort map from an opencv lens cal file
int mcv_init_undistort_map_file(mcv_undistort_map_t* map, const char* file, float zoom);



// initialize an undistort map from an opencv floating point bilinear map
// this is used under the hood for the stereo maps but can be used for any
// opencv bilinear map
int mcv_init_undistort_map_from_cvmat(int w, int h, mcv_undistort_map_t* map, float* cvmat);

// initialize two undistort maps for a stereo pair
int mcv_init_stereo_undistort_maps(const char* intrinsics_file, const char* extrinsics_file,
							mcv_undistort_map_t* l, mcv_undistort_map_t* r, float* baseline_m);

// undistort an image. Make sure memory for output is already allocated
int mcv_undistort_image(const uint8_t* input, uint8_t* output, mcv_undistort_map_t* map);

// undistort an rgb image. Make sure memory for output is already allocated
int mcv_undistort_rgb_image(const uint8_t* input, uint8_t* output, mcv_undistort_map_t* map);

// spawns off a thread to perform the undistort, wait for the thread to join
// to use the output data
int mcv_undistort_image_async(const uint8_t* input, uint8_t* output, mcv_undistort_map_t* map, pthread_t* thread);

int mcv_undistort_map_clean(mcv_undistort_map_t* map);



#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_UNDISTORT_CPU_H
