/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @headerfile  undistort_cvp.h <modalcv/undistort_cvp.h>
 *
 * @brief       Functions for undistorting and rectifying images
 *              using the CVP hardware.
 * 
 *              See the voxl-test-undistort.cpp example for use case.
 * 
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2023
 */

#ifndef MODALCV_UNDISTORT_CVP_H
#define MODALCV_UNDISTORT_CVP_H


#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <modal_pipe.h>

/**
 * @brief An object representing a generic handle for undistortion in modalcv
 * using CVP hardware. 
 */
typedef void* mcv_cvp_undistort_handle;

/**
 * @brief Different warp methods: grid, perspective, or perspective-grid
 * 
 * Note: Currently MCV_WARP_PERSPECTIVE and WARP_PERSPECTIVE_GRID are unused,
 * on the roadmap for future functionality
 */
typedef enum
{
    MCV_WARP_GRID = 0x1,
    MCV_WARP_PERSPECTIVE = 0x2,
    MCV_WARP_PERSPECTIVE_GRID = 0x3
} mcvWarpMode;

/**
 * @brief Different border modes: constant or replicate
 * 
 * Note: Currently MCV_BORDER_REPLICATE is unused but on the roadmap for future
 * functionality
 */
typedef enum
{
    MCV_BORDER_CONSTANT = 0x1,
    MCV_BORDER_REPLICATE = 0x2
} mcvBorderMode;

/**
 * @brief General structure for an undistortion grid
 */
typedef struct mcv_cvp_undistort_grid_t {
    double* x;
    double* y;
} mcv_cvp_undistort_grid_t;

/**
 * @brief Configuration info for the undistort process
 */
typedef struct mcv_undistort_config_t{
    int width;  
    int height; 
    int image_mode; 
    mcv_cvp_undistort_grid_t* undistort_grid;

    /* Unused */
    mcvWarpMode warp_mode;       ///< currently unused
    mcvBorderMode border_mode;   ///< currently unused
    float transform[9];          ///< currently unused

} mcv_undistort_config_t;

/**
 * @brief Helper function to create and allocate an undistort grid
 * 
 * @param w 
 * @param h 
 * @return mcv_cvp_undistort_grid_t* 
 */
mcv_cvp_undistort_grid_t* mcv_cvp_create_undistort_grid(int w, int h);

/**
 * @brief Helper function to tear down an undistort grid
 * 
 * @param undistort_grid 
 * @return int 
 */
int mcv_cvp_delete_grid(mcv_cvp_undistort_grid_t* undistort_grid);

/**
 * @brief Initializes the undistort process using the specified configuration
 * 
 * Most of this initialization involves configuring various memory buffers
 * for proprietary hardware. However, internally a calibration routine
 * is also performed. This looks at the intrinsics file passed in the config
 * variable and uses it to increase the accuracy of the undistort result
 * 
 * @param config A populated undistort configuration object
 * @return The generated undistort handle if successful, NULL otherwise
 */
mcv_cvp_undistort_handle mcv_cvp_undistort_init(mcv_undistort_config_t config);

/**
 * @brief De-initializes the undistort process 
 * 
 * @param handle The handle generated from mcv_undistort_init
 * @return 0 on success or -1 on failure
 */
int mcv_cvp_undistort_deinit(mcv_cvp_undistort_handle handle);

/**
 * @brief Executes the undistort process for a provided input image.
 * 
 * @param handle The undistort handle created via mcv_cvp_undistort_init
 * @param inImg The input image for undistorting
 * @param outImg The output image object
 * @return 0 on success or -1 on failure
 */
int mcv_cvp_undistort_process(
    mcv_cvp_undistort_handle handle,
    uint8_t* inImg,
    uint8_t* outImg
);

#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_UNDISTORT_CVP_H
