/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/stat.h>	// for mkdir
#include <sys/types.h>	// for mode_t in mkdir
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <iomanip>
#include <cvpDcm.h>

#include <map>
#include <vector>
#include <array>

#include <modalcv.h>
#include <modalcv/io.hpp>
#include <modalcv/dcm.h>

#include <fastcv.h>

#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/features2d.hpp>


// input files
#define TEST_IMAGE_1		"/usr/share/libmodalcv/ov9782/left.jpg"
#define TEST_IMAGE_2		"/usr/share/libmodalcv/ov9782/right.jpg"

// output files
#define OUTPUT_FOLDER	"/data/libmodalcv/dcm/"
#define OUTPUT_FILE		"/data/libmodalcv/dcm.png"

#define LOOPS 100


static int64_t _apps_time_monotonic_ns()
{
    struct timespec ts;
    if(clock_gettime(CLOCK_MONOTONIC, &ts)){
        fprintf(stderr,"ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec*1000000000 + (int64_t)ts.tv_nsec;
}

void printHexCodes(const void* data, int n) {
    const unsigned char* bytes = static_cast<const unsigned char*>(data);

    // Iterate over the number of codes
    for (int code_index = 0; code_index < n; ++code_index) {
        std::cout << "Code " << code_index + 1 << ": ";

        // Print 32 bytes for each code in hex format
        for (int i = 0; i < 32; ++i) {
            // Print each byte as two hexadecimal digits
            std::cout << std::hex << std::setw(2) << std::setfill('0') << static_cast<int>(bytes[code_index * 32 + i]);
        }

        std::cout << std::endl;  // End the current line
    }

    std::cout << std::dec;  // Reset the formatting to decimal
}


typedef struct feature_t {
    int x;
    int y;
    int id;
    int age;
    int octave;
    bool matched;
    unsigned char descriptor[32];
} feature_t;

std::vector<feature_t> feat_db;


// Custom comparator for char[32]
struct chr_arr_cmp {
    bool operator()(const char lhs[32], const char rhs[32]) const {
        return std::memcmp(lhs, rhs, 32) < 0;
    }
};

std::map<std::array<char, 32>, feature_t> feature_map[3];
// std::unordered_map<std::array<char, 32>, feature_t> feature_map;

typedef enum octave_num_t{
    OCT_0 = 0,
    OCT_1 = 1,
    OCT_2 = 2,
    N_OCTAVES
} octave_num_t;

// typedef enum var_state_t{
//     PREV = 0,
//     CURRENT = 1,
//     N_VAR_STATES
// } var_state_t;


int main(int argc, char* argv[]) {
    // mcv_en_debug_prints(0);
    
	int w, h;
    std::vector<std::string> paths;

    // args checks, makes sure path is good
    if (init_modalcv_example(argc, argv, paths, w, h)) {
        fprintf(stderr, "ERROR, failed to initialize example\n");
        return -1;
    }

    // create dir for img outputs
	if(system("rm -f /data/libmodalcv/dcm/*"))
	{
		fprintf(stderr, "Warning, failed to delete old files from %s\n", OUTPUT_FOLDER);
		return -1;
	}
    if(mkdir(OUTPUT_FOLDER, S_IRWXU) && errno!=EEXIST){
        fprintf(stderr, "failed to make folder\n");
        return -1;
    }

    /********* PYR FPX SETUP *********/

    // init pyramid extractor config
    mcv_cvp_pyr_fpx_config_t pyr_fpx_config;
    pyr_fpx_config.width = w;
    pyr_fpx_config.height = h;
    pyr_fpx_config.n_octaves = 3;
    pyr_fpx_config.n_scales_per_octave = 1;
    pyr_fpx_config.mode = MCV_CVP_FPX_ZONE;
    pyr_fpx_config.nms_mode = MCV_CVP_FPX_5_TAP_NMS;
    pyr_fpx_config.score_threshold = 80;
    pyr_fpx_config.robustness = 10;  // 0-127, default 10
    pyr_fpx_config.zone_dims[0] = 16;
    pyr_fpx_config.zone_dims[1] = 8;
    pyr_fpx_config.zone_dims[2] = 4;
    
    for (int i = 0; i < MCV_CVP_MAX_PYRAMID_OCTAVES; i++) {
        pyr_fpx_config.enable_fpx[i] = 1;
    }

    // create a pyr_fpx handle from our configuration
    mcv_cvp_pyr_fpx_handle pyr_fpx_handle = mcv_cvp_pyr_fpx_init(pyr_fpx_config);
    if (NULL == pyr_fpx_handle) {
        fprintf(stderr, "ERROR creating pyr fpx handle\n");
    }

	// creates output structure
	mcv_cvp_pyr_fpx_t output_pyr_fpx;
    if (mcv_cvp_create_pyr_fpx(pyr_fpx_handle, &output_pyr_fpx)) {
        fprintf(stderr, "ERROR creating feature extraction pyramid\n");
        return -1;
    }

    /********* DCM SETUP *********/

    // configure DCM and get handles
    mcv_cvp_dcm_handle dcm_handles[N_OCTAVES];
    
    // create an output object
    mcv_cvp_dcm_output_t dcm_outputs[N_OCTAVES][N_VAR_STATES];
    std::vector<int> cur_hamming_match_idx[N_OCTAVES];

    for (int i = 0; i < N_OCTAVES; i++) {
        mcv_cvp_dcm_config_t dcm_config;
        dcm_config.w = w >> i;
        dcm_config.h = h >> i;

        dcm_handles[i] = mcv_cvp_dcm_init(dcm_config);
        if(NULL == dcm_handles[i]) {
            fprintf(stderr, "ERROR getting DCM handle\n");
            return -1;
        }
        if(mcv_cvp_create_dcm_output(dcm_handles[i], &dcm_outputs[i][CURRENT])) {
            fprintf(stderr, "ERROR creating DCM output object\n");
            return -1;
        }
        if(mcv_cvp_create_dcm_output(dcm_handles[i], &dcm_outputs[i][PREV])) {
            fprintf(stderr, "ERROR creating DCM output object\n");
            return -1;
        }   
    }

    int id_count = 0;

    /********* DO THE STUFF *********/
    bool first_frame = true; // must start as true

    bool do_cvp_matching = true;
    bool do_hamming_matching = false;

    std::vector<mcv_cvp_fpx_feature_t> prev_fpx[N_OCTAVES];

    uint8_t* prev_img[N_OCTAVES];

    for (size_t i = 0; i < paths.size(); i++) {
        // printf("paths: %s\n", paths[i].c_str());

        cv::Mat cur_img = cv::imread(paths[i], cv::IMREAD_GRAYSCALE);

        if (cur_img.type()!= CV_8UC1) {
            fprintf(stderr, "ERROR loading %s\n", paths[i].c_str());
            return -1;
        }

        // add preprocess blur before doing pyramid feature extraction
        cv::GaussianBlur(cur_img, cur_img,cv::Size(3, 3), 0);
        mcv_cvp_pyr_fpx_process(pyr_fpx_handle, cur_img.data, &output_pyr_fpx);

        mcv_cvp_fpx_t fpx[N_OCTAVES];
        mcv_cvp_fpx_t prev_fpx_copy[N_OCTAVES];

        for (int t = 0; t < N_OCTAVES; t++) {
            fpx[t].features = output_pyr_fpx.fpx[t];
            fpx[t].n_points = output_pyr_fpx.n_points[t];

            prev_fpx_copy[t].features = prev_fpx[t].data();
            prev_fpx_copy[t].n_points = prev_fpx[t].size();
        }

        if (first_frame) {
            for (int oct = 0; oct < N_OCTAVES; oct++) {
                mcv_cvp_dcm_process(dcm_handles[oct], output_pyr_fpx.image_pyramids[oct], &fpx[oct], NULL, 0, &dcm_outputs[oct][CURRENT]);
                
                if (dcm_outputs[oct][CURRENT].n_descriptors && output_pyr_fpx.n_points[oct]) {
                    printf("got features for oct %d\n", oct);
                    first_frame = false;

                    prev_img[oct] = (uint8_t*)malloc(output_pyr_fpx.width[oct] * output_pyr_fpx.height[oct] * sizeof(uint8_t));
                    
                    for (int j = 0; j < fpx[oct].n_points; j++) {
                        feature_t feat;
                        feat.x = fpx[oct].features[j].x;
                        feat.y = fpx[oct].features[j].y;
                        feat.id = id_count++;
                        feat.matched = false;
                        feat.octave = oct;
                        feat.age = 1;

                        void* desc_src = static_cast<unsigned char*>(dcm_outputs[oct][CURRENT].descriptors) + j * 32;
                        memcpy(feat.descriptor, desc_src, 32);

                        std::array<char, 32> arr_array;
                        std::copy(std::begin(feat.descriptor), std::end(feat.descriptor), arr_array.begin());

                        feature_map[oct][arr_array] = feat;
                        feat_db.push_back(feat);
                    }
                } else {
                    printf("NOT GETTING FEATURES\n");
                }
            }

        } else {
            cv::Mat concatenated;
            cv::Mat cur_img_copy = cur_img;
            cv::cvtColor(cur_img_copy, cur_img_copy, cv::COLOR_GRAY2RGB);
            cv::Mat prev_img_copy = cv::Mat(output_pyr_fpx.height[0], 
                                            output_pyr_fpx.width[0], CV_8UC1, prev_img[0]);
            cv::cvtColor(prev_img_copy, prev_img_copy, cv::COLOR_GRAY2RGB);
            cv::hconcat(prev_img_copy, cur_img_copy, concatenated);

            cur_img_copy = cv::Mat(output_pyr_fpx.height[1], 
                                   output_pyr_fpx.width[1], CV_8UC1, output_pyr_fpx.image_pyramids[1]);
            prev_img_copy = cv::Mat(output_pyr_fpx.height[1], 
                                    output_pyr_fpx.width[1], CV_8UC1, prev_img[1]);

            cv::Mat h_concatenated;
            cv::resize(cur_img_copy, cur_img_copy, 
                        cv::Size(output_pyr_fpx.width[0], output_pyr_fpx.height[0]),
                        0, 0, cv::INTER_NEAREST);
            cv::resize(prev_img_copy, prev_img_copy, 
                        cv::Size(output_pyr_fpx.width[0], output_pyr_fpx.height[0]),
                        0, 0, cv::INTER_NEAREST);
            cv::cvtColor(prev_img_copy, prev_img_copy, cv::COLOR_GRAY2RGB);
            cv::cvtColor(cur_img_copy, cur_img_copy, cv::COLOR_GRAY2RGB);
            cv::hconcat(prev_img_copy, cur_img_copy, h_concatenated);

            cv::vconcat(concatenated, h_concatenated, concatenated);

            cur_img_copy = cv::Mat(output_pyr_fpx.height[2], 
                                   output_pyr_fpx.width[2], CV_8UC1, output_pyr_fpx.image_pyramids[2]);
            prev_img_copy = cv::Mat(output_pyr_fpx.height[2], 
                                    output_pyr_fpx.width[2], CV_8UC1, prev_img[2]);
            cv::resize(cur_img_copy, cur_img_copy, 
                        cv::Size(output_pyr_fpx.width[0], output_pyr_fpx.height[0]),
                        0, 0, cv::INTER_NEAREST);
            cv::resize(prev_img_copy, prev_img_copy,
                        cv::Size(output_pyr_fpx.width[0], output_pyr_fpx.height[0]),
                        0, 0, cv::INTER_NEAREST);
            cv::cvtColor(prev_img_copy, prev_img_copy, cv::COLOR_GRAY2RGB);
            cv::cvtColor(cur_img_copy, cur_img_copy, cv::COLOR_GRAY2RGB);
            cv::hconcat(prev_img_copy, cur_img_copy, h_concatenated);
            cv::vconcat(concatenated, h_concatenated, concatenated);

            int64_t iter_start = _apps_time_monotonic_ns();
            int64_t end;

            for (int oct = 0; oct < N_OCTAVES; oct++) {
                // doing dcm process
                int64_t start = _apps_time_monotonic_ns();
                mcv_cvp_dcm_process(dcm_handles[oct], output_pyr_fpx.image_pyramids[oct], &fpx[oct], NULL, 0, &dcm_outputs[oct][CURRENT]);

                if (do_cvp_matching) {
                    mcv_cvp_dcm_process(dcm_handles[oct], prev_img[oct], &prev_fpx_copy[oct], 
                                        dcm_outputs[oct][CURRENT].descriptors, dcm_outputs[oct][CURRENT].n_descriptors, 
                                        &dcm_outputs[oct][PREV]);
                }

                // std::vector<int> cur_hamming_match_idx;
                const uint8_t zero_descriptor[32] = {0}; 

                if (do_hamming_matching) {
                    // calc hamming distance for each
                    for(int prev_i = 0; prev_i < dcm_outputs[oct][PREV].n_descriptors; prev_i++) {
                        uint32_t best_dist = UINT32_MAX;
                        uint32_t best_idx = -1;

                        for (int cur_i = 0; cur_i < dcm_outputs[oct][CURRENT].n_descriptors; cur_i++) {

                            uint32_t dist = fcvHammingDistance32x1u8(
                                                (const uint8_t*)dcm_outputs[oct][CURRENT].descriptors + 32 * cur_i,
                                                (const uint8_t*)dcm_outputs[oct][PREV].descriptors    + 32 * prev_i);

                            if (dist < best_dist) {
                                best_dist = dist;
                                best_idx = cur_i;
                            }
                        }

                        printf("best match:(%d, %d) --> (%d, %d), dist: %d\n", 
                                prev_fpx_copy[oct].features[prev_i].x, prev_fpx_copy[oct].features[prev_i].y,
                                fpx[oct].features[best_idx].x, fpx[oct].features[best_idx].y, best_dist);

                        printf("previous\n");
                        printHexCodes(dcm_outputs[oct][PREV].descriptors + 32 * prev_i, 1);

                        printf("current\n");
                        printHexCodes(dcm_outputs[oct][CURRENT].descriptors + 32 * best_idx, 1);

                        cur_hamming_match_idx[oct].push_back(best_idx);
                    }
                }

                end = _apps_time_monotonic_ns();
                printf("dcm process ::    %6.2fms, oct: %d\n", (double)(end - start) / 1000000, oct);
                // printf("path: %s\n", paths[i].c_str());

                // DOING RANSAC ON THE DCM MATCHES
                const float inlier_threshold = 0.02f * pow(2, oct);
                const float success_probability = 0.9f;
                const float outlier_ratio = 0.35f;
                const int min_point_for_model = 1;

                // number of iterations to find ideal model
                int N = static_cast<int>(std::log(1 - success_probability) / std::log(1 - std::pow(1 - outlier_ratio, min_point_for_model)));

                float best_slope[N_OCTAVES];
                int max_inliers[N_OCTAVES];

                int inlier_count = 0;

                if (do_cvp_matching || do_hamming_matching) {

                    int n_matches;
                    if (do_cvp_matching) {
                        n_matches = dcm_outputs[oct][PREV].n_matches;
                    } else if (do_hamming_matching) {
                        n_matches = prev_fpx[oct].size();
                    }

                    for(int n = 0; n < N; n++) {
                        int idx = std::rand() % n_matches;

                        int cur_img_x, cur_img_y;

                        if (do_cvp_matching) {
                            cur_img_x = fpx[oct].features[dcm_outputs[oct][PREV].matches[idx].index].x + w;
                            cur_img_y = fpx[oct].features[dcm_outputs[oct][PREV].matches[idx].index].y;
                        } else if (do_hamming_matching) {
                            cur_img_x = fpx[oct].features[cur_hamming_match_idx[oct][idx]].x + w;
                            cur_img_y = fpx[oct].features[cur_hamming_match_idx[oct][idx]].y;
                        }

                        int prev_img_x = prev_fpx[oct][idx].x;
                        int prev_img_y = prev_fpx[oct][idx].y;

                        float slope = (cur_img_y - prev_img_y) / static_cast<float>(cur_img_x - prev_img_x);

                        inlier_count = 0;
                        for (int j = 0; j < n_matches; ++j) {
                            if (j != idx) {

                                int cur_img_x, cur_img_y;
                                if (do_cvp_matching) {
                                    cur_img_x = fpx[oct].features[dcm_outputs[oct][PREV].matches[j].index].x + w; // >> oct
                                    cur_img_y = fpx[oct].features[dcm_outputs[oct][PREV].matches[j].index].y;
                                } else if (do_hamming_matching) {
                                    cur_img_x = fpx[oct].features[cur_hamming_match_idx[oct][j]].x + w; // >> oct
                                    cur_img_y = fpx[oct].features[cur_hamming_match_idx[oct][j]].y;
                                }

                                int prev_img_x = prev_fpx[oct][j].x;
                                int prev_img_y = prev_fpx[oct][j].y;

                                float cur_slope = (cur_img_y - prev_img_y) / static_cast<float>(cur_img_x - prev_img_x);

                                if (std::abs(cur_slope - slope) < inlier_threshold) {
                                    ++inlier_count;
                                }
                            }
                        }

                        if (inlier_count > max_inliers[oct]) {
                            max_inliers[oct] = inlier_count;
                            best_slope[oct] = slope;
                        }
                    }

                    // RANSAC FINISHED
                    for (int j = 0; j < n_matches; ++j) {
                        int cur_img_x, cur_img_y;

                        if (do_cvp_matching) {
                            cur_img_x = fpx[oct].features[dcm_outputs[oct][PREV].matches[j].index].x * pow(2, oct) + w;
                            cur_img_y = fpx[oct].features[dcm_outputs[oct][PREV].matches[j].index].y * pow(2, oct) + (h * oct);
                        } else if (do_hamming_matching) {
                            cur_img_x = fpx[oct].features[cur_hamming_match_idx[oct][j]].x * pow(2, oct) + w;
                            cur_img_y = fpx[oct].features[cur_hamming_match_idx[oct][j]].y * pow(2, oct) + (h * oct);
                        }

                        int prev_img_x = prev_fpx[oct][j].x * pow(2, oct);
                        int prev_img_y = prev_fpx[oct][j].y * pow(2, oct) + (h * oct);

                        float slope = (cur_img_y - prev_img_y) / static_cast<float>(cur_img_x - prev_img_x);

                        std::array<char, 32> descriptor_arr;
                        void* desc_src = static_cast<unsigned char*>(dcm_outputs[oct][PREV].descriptors) + j * 32;
                        std::memcpy(descriptor_arr.data(), desc_src, 32);

                        // this is the new descriptor for the same feature that was found
                        int matched_idx;
                        if (do_cvp_matching) {
                            matched_idx = dcm_outputs[oct][PREV].matches[j].index;
                        } else if (do_hamming_matching) {
                            matched_idx = cur_hamming_match_idx[oct][j];
                        }
                        void* match_src = static_cast<unsigned char*>(dcm_outputs[oct][CURRENT].descriptors) + matched_idx * 32;
                        std::array<char, 32> match_arr;
                        std::memcpy(match_arr.data(), match_src, 32);

                        cv::Scalar color;
                        if (std::abs(best_slope[oct] - slope) < inlier_threshold) {
                            int x = 0;

                            auto it = feature_map[oct].find(descriptor_arr);
                            if (it != feature_map[oct].end()) {
                                
                                // getting the index that maps to the new descriptor for the same feature

                                // creating a temp feature to copy data from last descriptor
                                feature_t temp = feature_map[oct][descriptor_arr];
                                x = temp.id;
                                temp.age++;
                                memcpy(temp.descriptor, match_src, 32);

                                // erase old descriptor key for feature
                                feature_map[oct].erase(descriptor_arr);

                                // add new descriptor key for feature
                                feature_map[oct][match_arr] = temp;
                            } else {
                                // old feature descriptor not matched, remove it from the map
                                feature_map[oct].erase(descriptor_arr);
                            }

                            color = cv::Scalar((x * 50) % 256, (x * 80) % 256, (x * 120) % 256);
                        } else {
                            // continue;
                        }

                        cv::line(concatenated, cv::Point(prev_img_x, prev_img_y), cv::Point(cur_img_x, cur_img_y), color, 2, cv::LINE_AA);
                        cv::circle(concatenated, cv::Point(cur_img_x, cur_img_y), 2, color, cv::FILLED);
                        cv::circle(concatenated, cv::Point(prev_img_x, prev_img_y), 2, color, cv::FILLED);
                    }
                }

                
                // all unmatched new descriptors get added as new features in the map
                for (int j = 0; j < dcm_outputs[oct][CURRENT].n_descriptors; j++) {

                    // cehck that the descriptor is not present in the 
                    std::array<char, 32> descriptor_arr;
                    void* desc_src = static_cast<unsigned char*>(dcm_outputs[oct][CURRENT].descriptors) + j * 32;
                    std::memcpy(descriptor_arr.data(), desc_src, 32);

                    auto it = feature_map[oct].find(descriptor_arr);
                    if (it == feature_map[oct].end()) { 
                        feature_t temp;
                        feature_t feat;
                        feat.x = fpx[oct].features[j].x;
                        feat.y = fpx[oct].features[j].y;
                        feat.id = id_count++;
                        feat.matched = false;
                        feat.age = 1;
                        feat.octave = oct;

                        feature_map[oct][descriptor_arr] = feat;
                    }
                }
                // printf("size of map: %d\n", feature_map.size());
            }
            end = _apps_time_monotonic_ns();
            printf("Iter Time ::    %6.2fms\n", (double)(end - iter_start) / 1000000);


            if (do_cvp_matching || do_hamming_matching) {
                std::stringstream ss;
                ss << std::setw(5) << std::setfill('0') << i;
                std::string output_path = OUTPUT_FOLDER + ss.str() + ".png";
                printf("%s\n", output_path.c_str());

                cv::imwrite(output_path, concatenated);
            }
        }

        for (int oct = 0; oct < N_OCTAVES; oct++) {
            // do a memcpy for the images
            memcpy(prev_img[oct], output_pyr_fpx.image_pyramids[oct], 
                   output_pyr_fpx.width[oct] * output_pyr_fpx.height[oct]);
                
            dcm_outputs[oct][PREV].n_descriptors = dcm_outputs[oct][CURRENT].n_descriptors;
            if (do_cvp_matching) {
                dcm_outputs[oct][PREV].n_matches = dcm_outputs[oct][CURRENT].n_matches;
            } else if (do_hamming_matching) {
                
            }

            memcpy(dcm_outputs[oct][PREV].descriptors, dcm_outputs[oct][CURRENT].descriptors,
                   dcm_outputs[oct][CURRENT].n_descriptors * 32);

            prev_fpx[oct].clear();
            for (int j = 0; j < fpx[oct].n_points; j++) {
                prev_fpx[oct].push_back(fpx[oct].features[j]);
            }
        }
    }



    /********* CLEANUP *********/

    // cleanup for pyramid extraction
    if(mcv_cvp_destroy_pyr_fpx(pyr_fpx_handle, &output_pyr_fpx)) {
		fprintf(stderr, "ERROR destorying pyramid object\n");
		return -1;
	}
	if(mcv_cvp_pyr_fpx_deinit(pyr_fpx_handle)) {
		fprintf(stderr, "ERROR deinitializing pyramid\n");
		return -1;
	}

    // dcm cleanup
    for (int i = 0; i < N_OCTAVES; i++) {
        if(mcv_cvp_destroy_dcm_output(dcm_handles[i], &dcm_outputs[i][CURRENT])) {
            fprintf(stderr, "ERROR destroying dcm output\n");
            return -1;
        }
        if(mcv_cvp_destroy_dcm_output(dcm_handles[i], &dcm_outputs[i][PREV])) {
            fprintf(stderr, "ERROR destroying dcm output\n");
            return -1;
        }

        if(mcv_cvp_dcm_deinit(dcm_handles[i])) {
            fprintf(stderr, "ERROR deinitializing dcm\n");
            return -1;
        }
    }

    return 0;
}
