/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/stat.h>	// for mkdir
#include <sys/types.h>	// for mode_t in mkdir
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <modalcv.h>

#include <opencv2/imgcodecs.hpp>
#include <opencv2/core/mat.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/imgproc.hpp>

using namespace cv;

// input files

// 640x480
// #define TEST_IMAGE_L        "/usr/share/libmodalcv/dfs/left.png"
// #define TEST_IMAGE_R        "/usr/share/libmodalcv/dfs/right.png"
// #define INTRINSICS_FILE     "/usr/share/libmodalcv/dfs/opencv_stereo_intrinsics.yml"
// #define EXTRINSICS_FILE     "/usr/share/libmodalcv/dfs/opencv_stereo_extrinsics.yml"

// 1280x800
#define TEST_IMAGE_L		"/usr/share/libmodalcv/dfs2/00900.png"
#define TEST_IMAGE_R		"/usr/share/libmodalcv/dfs2/00905.png"
#define INTRINSICS_FILE		"/usr/share/libmodalcv/dfs2/opencv_stereo_grey_intrinsics.yml"
#define EXTRINSICS_FILE		"/usr/share/libmodalcv/dfs2/opencv_stereo_grey_extrinsics.yml"

// output files
#define OUTPUT_FOLDER		"/data/libmodalcv/"

#define IMGL_FILT_FILE		"/data/libmodalcv/dfs_imgL_filt.png"
#define IMGR_FILT_FILE		"/data/libmodalcv/dfs_imgR_filt.png"
#define IMGL_RECT_FILE		"/data/libmodalcv/dfs_imgL_rect.png"
#define IMGR_RECT_FILE		"/data/libmodalcv/dfs_imgR_rect.png"
#define COMBINED_FILE		"/data/libmodalcv/dfs_combined.png"
#define DISPARITY_RAST_FILE	"/data/libmodalcv/dfs_disparity_raster.png"
#define DISPARITY_FILE		"/data/libmodalcv/dfs_disparity.png"
#define OCCLUSION_FILE		"/data/libmodalcv/dfs_occlusion.png"

#define MIN_DISPARITY 0
#define LOOPS 100

/**
 * @brief Different hardware methods to use for testing
 * 
 * MCV_CPU: Computes undistortion with CPU, then computes
 * DFS using CVP hardware.
 * 
 * MCV_CVP: Computes undistortion with CVP, then comptues
 * DFS using CVP hardware
 * 
 * MCV_CVP_PIPELINE: Computes undistortion and DFS using a CVP pipeline, 
 * passing pointers to CVP memory between the tasks
 * 
 */
typedef enum
{
    MCV_CPU = 0x1,
    MCV_CVP = 0x2,
    MCV_CVP_PIPELINE = 0x3
} mcv_dfs_hardware_spec;

/**
 * @brief Helper fn for getting the current clock time in nanoseconds
 * 
 * @return Time in ns
 */
static int64_t _apps_time_monotonic_ns()
{
	struct timespec ts;
	if(clock_gettime(CLOCK_MONOTONIC, &ts)){
		fprintf(stderr,"ERROR calling clock_gettime\n");
		return -1;
	}
	return (int64_t)ts.tv_sec*1000000000 + (int64_t)ts.tv_nsec;
}

/**
 * @brief Create directories for output files and delete any old files
 * 
 * @return 0 on success, -1 otherwise
 */
static int init_directories()
{
	// make output directory to save debug images to
	if(mkdir(OUTPUT_FOLDER, S_IRWXU) && errno!=EEXIST)
	{
		fprintf(stderr, "failed to make folder\n");
		return -1;
	}
	if(system("rm -f /data/libmodalcv/*"))
	{
		fprintf(stderr, "Warning, failed to delete old files from %s\n", OUTPUT_FOLDER);
		return -1;
	}
	return 0;
}

/**
 * @brief Loads and preprocesses the images
 * 
 * @param L ptr to output left image
 * @param R ptr to output right image
 * @return int 
 */
static int preprocess_images(Mat* L, Mat* R)
{
	Mat imgL, imgR;
	imgL = imread(TEST_IMAGE_L, IMREAD_GRAYSCALE);
	if(imgL.type()!=CV_8UC1){
		fprintf(stderr, "ERROR loading %s\n", TEST_IMAGE_L);
		return -1;
	}
	imgR = imread(TEST_IMAGE_R, IMREAD_GRAYSCALE);
	if(imgR.type()!=CV_8UC1){
		fprintf(stderr, "ERROR loading %s\n", TEST_IMAGE_R);
		return -1;
	}
	cv::setNumThreads(0); // for small images, opencv is typically faster without multithreading
	Mat imgL_tmp1, imgR_tmp1, imgL_tmp2, imgR_tmp2;
	Mat imgL_filt, imgR_filt;

	Size blursize = Size(13,13);
	GaussianBlur(imgL, imgL_filt, blursize, 0, 0, BORDER_DEFAULT);
	GaussianBlur(imgR, imgR_filt, blursize, 0, 0, BORDER_DEFAULT);

	// copy out
	*L = imgL_filt;
	*R = imgR_filt;
	return 0;
}
	
/**
 * @brief Configure pointcloud
 * 
 * @param pc_disp_table 
 * @return int 
 */
// static int config_pc(mcv_pc_disp_table_t* pc_disp_table)
// {
// 	// PointCloud Initialization
// 	mcv_pc_disp_config_t pc_config;
// 	pc_config.width         = w;
// 	pc_config.height        = h;
// 	pc_config.f             = map_l.fxrect;
// 	pc_config.cx            = roundf(map_l.cxrect);
// 	pc_config.cy            = roundf(map_l.cxrect);
// 	pc_config.baseline_m    = baseline_m;
// 	pc_config.skip_lines    = 3                // only do some of the points
// 	pc_config.min_disparity = MIN_DISPARITY;
// 	if(mcv_pc_disp_table_init(pc_config, pc_disp_table)) return -1;
// }

/**
 * @brief Populates DFS configuration object
 * 
 * @param dfs_config 
 * @param w 
 * @param h 
 * @return int 
 */
static int populate_dfs_config(mcv_dfs_config_t* dfs_config, int w, int h)
{
	dfs_config->width                    = w;
	dfs_config->height                   = h;
	dfs_config->n_disparity_levels            = 32;
	dfs_config->starting_disparity_offset         = MIN_DISPARITY;  // Valid range is [0, 254-nMaxDisparity]
	dfs_config->en_median_filter         = 1;              // enable median filtering
	dfs_config->en_occlusion_filling     = 0;              // don't fill in data to occluded areas
	dfs_config->occlusion_min_threshold  = 50;              // [0-100] nOcclusionMinThreshold <= nOcclusionThreshold
	dfs_config->occlusion_cost_threshold = 50;              // [0-100] nOcclusionMinThreshold <= nOcclusionThreshold
	return 0;
}

/**
 * @brief Populates undistortion configuration object
 * 
 * @param undistort_config 
 * @return int 
 */
static int populate_undistort_config(mcv_undistort_config_t* undistort_config_l, 
	mcv_undistort_config_t* undistort_config_r, mcv_stereo_params_t* stereo_params,
	int w, int h)
{

	// populate left/right configs with some values
	undistort_config_l->width       = w;
	undistort_config_l->height      = h;
	undistort_config_l->image_mode  = IMAGE_FORMAT_RAW8;
	undistort_config_l->warp_mode   = MCV_WARP_GRID;
	undistort_config_l->border_mode = MCV_BORDER_CONSTANT;
	undistort_config_l->undistort_grid = mcv_cvp_create_undistort_grid(w, h);

	undistort_config_r->width       = w;
	undistort_config_r->height      = h;
	undistort_config_r->image_mode  = IMAGE_FORMAT_RAW8;
	undistort_config_r->warp_mode   = MCV_WARP_GRID;
	undistort_config_r->border_mode = MCV_BORDER_CONSTANT;
	undistort_config_r->undistort_grid = mcv_cvp_create_undistort_grid(w, h);

	// use intrinsics/extrinsics to populate undistortion grids for L/R
	if(mcv_create_stereo_undistort_grid(INTRINSICS_FILE, EXTRINSICS_FILE, 
		stereo_params, undistort_config_l->undistort_grid, 
		undistort_config_r->undistort_grid))
	{
		fprintf(stderr, "ERROR creating undistortion grids for L/R");
	}
	return 0;
}

/**
 * @brief Undistorts the left and right images using the CPU
 * 
 * @param left_input 
 * @param right_input 
 * @param right_input 
 * @param right_output 
 * @param map_l 
 * @param map_r 
 * @return int 
 */
static int run_undistort_cpu(uint8_t* left_input, uint8_t* right_input, 
	uint8_t* left_output, uint8_t* right_output, mcv_undistort_map_t* map_l,
	mcv_undistort_map_t* map_r)
{

	int64_t t0 = _apps_time_monotonic_ns();
	for(int i=0; i<LOOPS; i++)
	{
		if(mcv_undistort_image(left_input, left_output,  map_l))
		{
			fprintf(stderr, "ERROR in computing undistortion for left image\n");
			return -1;
		}
		if(mcv_undistort_image(right_input, right_output, map_r))
		{
			fprintf(stderr, "ERROR in computing undistortion for right image\n");
			return -1;
		}
	}
	int64_t t1 = _apps_time_monotonic_ns();
	int64_t t_per_loop_ns = (t1-t0)/LOOPS;
	printf("%6.2fms   CPU Undistort Time\n", ((double)(t_per_loop_ns))/1000000.0);
	
	// cleanup
	return 0;
}

static int run_undistort_cvp(mcv_undistort_config_t undistort_config_l,
	mcv_undistort_config_t undistort_config_r, uint8_t* left_input,
	uint8_t* right_input, uint8_t* left_output, uint8_t* right_output)
{
	// use configuration to get an undistort handle for L, then cp to R
	mcv_cvp_undistort_handle undistort_handle_l, undistort_handle_r;
	undistort_handle_l = mcv_cvp_undistort_init(undistort_config_l);
	undistort_handle_r = mcv_cvp_undistort_init(undistort_config_r);
	if(undistort_handle_l == NULL)
	{
		printf("Error in configuring undistort handle\n");
		return -1;
	}
	if(undistort_handle_r == NULL)
	{
		printf("Error in configuring undistort handle\n");
		return -1;
	}

	// compute the undistortion for left and right, writing to the output buf
	int64_t t0 = _apps_time_monotonic_ns();
	for(int i=0; i<LOOPS; i++)
	{
		if(mcv_cvp_undistort_process(undistort_handle_l, left_input, left_output))
		{
			printf("Error in completing CVP undistortion for the left image\n");
			return -1;
		}
		if(mcv_cvp_undistort_process(undistort_handle_r, right_input, right_output))
		{
			printf("Error in completing CVP undistortion for the right image\n");
			return -1;
		}
	}
	int64_t t1 = _apps_time_monotonic_ns();
	int64_t t_per_loop_ns = (t1-t0)/LOOPS;
	printf("%6.2fms   CVP Undistort Time\n", ((double)(t_per_loop_ns))/1000000.0);


	// cleanup
	if(mcv_cvp_undistort_deinit(undistort_handle_l))
	{
		printf("Error: Failed to deinit undistort handle left\n");
		return -1;
	}
	if(mcv_cvp_undistort_deinit(undistort_handle_r))
	{
		printf("Error: Failed to deinit undistort handle right\n");
		return -1;
	}
	return 0;
}

/**
 * @brief Computes DFS for a left and right image using CVP hardware
 * 
 * @param left 
 * @param right 
 * @param disparity 
 * @param occlusion_mask 
 * @return int 
 */
static int run_dfs_cvp(mcv_cvp_dfs_handle handle, uint8_t* left, uint8_t* right, uint8_t* disparity, 
	uint8_t* occlusion_mask)
{
	int64_t t0 = _apps_time_monotonic_ns();
	for (int i=0; i<LOOPS; i++)
	{
		if(mcv_cvp_dfs_process(handle, left, right, disparity, occlusion_mask))
		{
			fprintf(stderr, "ERROR in computing DFS using CPU\n");
			return -1;
		}
	}
	int64_t t1 = _apps_time_monotonic_ns();
	int64_t t_per_loop_ns = (t1-t0)/LOOPS;
	printf("%6.2fms   CVP DFS Time\n", ((double)(t_per_loop_ns))/1000000.0);
	return 0;
}

static int run_undistort_dfs_pipeline(
	mcv_undistort_config_t undistort_config_l,
	mcv_undistort_config_t undistort_config_r, mcv_dfs_config_t dfs_config,
	uint8_t* left, uint8_t* right, uint8_t* disparity, uint8_t* occlusion_mask)
{
	// use both Undistort and DFS configs to get pipeline handle
	mcv_cvp_undistort_and_dfs_handle handle;
	handle = mcv_cvp_undistort_and_dfs_init(undistort_config_l, undistort_config_r, dfs_config);
	if(NULL == handle)
	{
		fprintf(stderr, "ERROR: Undistort+DFS handle is NULL in %s\n", __FUNCTION__);
		return -1;
	}

	// run Undistort+DFS pipeline
	int64_t t0 = _apps_time_monotonic_ns();
	for (int i=0; i<LOOPS; i++)
	{
		if(mcv_cvp_undistort_and_dfs_process(handle, left, right, disparity, occlusion_mask))
		{
			fprintf(stderr, "ERROR computing Undistort+DFS process in %s\n", __FUNCTION__);
			return -1;
		}
	}
	int64_t t1 = _apps_time_monotonic_ns();
	int64_t t_per_loop_ns = (t1-t0)/LOOPS;
	printf("%6.2fms   CVP Pipeline Time\n", ((double)(t_per_loop_ns))/1000000.0);

	// deinit
	if(mcv_cvp_undistort_and_dfs_deinit(handle))
	{
		fprintf(stderr, "ERROR deinitializing Undistort+DFS pipeline\n");
		return -1;
	}

	return 0;
}


/**
 * @brief Runs a single DFS process on the specified piece of hardware
 * 
 * @return int 
 */
static int run_test(mcv_dfs_hardware_spec hardware_spec)
{
	// initialize maps
	mcv_undistort_map_t map_l = MCV_UNDISTORT_MAP_INITIALIZER;
	mcv_undistort_map_t map_r = MCV_UNDISTORT_MAP_INITIALIZER;
	float baseline_m;
	if(mcv_init_stereo_undistort_maps(INTRINSICS_FILE, EXTRINSICS_FILE, &map_l, &map_r, &baseline_m))
	{
		fprintf(stderr, "Error initializing stereo undistort maps in %s\n", __FUNCTION__);
		return -1;
	}
	int w = map_l.w;
	int h = map_l.h;

	// load and preprocess images
	cv::Mat imgL, imgR;
	preprocess_images(&imgL, &imgR);

	// populate configurations for Undistort and DFS
	mcv_stereo_params_t stereo_params;
	mcv_undistort_config_t undistort_config_l, undistort_config_r;
	mcv_dfs_config_t dfs_config;
	if(populate_undistort_config(&undistort_config_l, &undistort_config_r, 
		&stereo_params, w, h))
	{
		fprintf(stderr, "ERROR populating Undistort configs  in %s\n", __FUNCTION__);
		return -1;
	}
	if(populate_dfs_config(&dfs_config, w, h))
	{
		fprintf(stderr, "ERROR populating DFS config in %s\n", __FUNCTION__);
		return -1;
	}

	// create DFS handle 
	mcv_cvp_dfs_handle dfs_handle = mcv_cvp_dfs_init(dfs_config);
	if(dfs_handle == NULL)
	{
		fprintf(stderr, "ERROR initializing DFS, handle is NULL\n");
		return -1;
	}

	// malloc for outputs
	uint8_t* imgL_rect      = (uint8_t*)malloc(w*h);
	uint8_t* imgR_rect      = (uint8_t*)malloc(w*h);
	uint8_t* disparity_rast = (uint8_t*)malloc(w*h);
	uint8_t* occlusion_rast = (uint8_t*)malloc(w*h);
	uint8_t* disparity_lin  = (uint8_t*)malloc(w*h);
	uint8_t* occlusion_lin  = (uint8_t*)malloc(w*h);

	// run the undistort + DFS process based on the provided spec
	switch(hardware_spec)
	{
		case MCV_CPU:
			printf("----- TEST: CPU            -----\n");
			if(run_undistort_cpu(imgL.data, imgR.data, imgL_rect, imgR_rect, &map_l, &map_r))
			{
				fprintf(stderr, "ERROR in computing undistortion on CPU\n");
				return -1;
			}
			if(run_dfs_cvp(dfs_handle, imgL_rect, imgR_rect, disparity_rast, occlusion_rast))
			{
				fprintf(stderr, "ERROR in computing DFS on CVP\n");
				return -1;
			}
			break;
		case MCV_CVP:
			printf("----- TEST: CVP            -----\n");
			if(run_undistort_cvp(undistort_config_l, undistort_config_r, 
				imgL.data, imgR.data, imgL_rect, imgR_rect))
			{
				fprintf(stderr, "ERROR in computing undistortion on CVP\n");
				return -1;
			}
			if(run_dfs_cvp(dfs_handle, imgL_rect, imgR_rect, disparity_rast, occlusion_rast))
			{
				fprintf(stderr, "ERROR in computing undistortion on CVP\n");
				return -1;
			}
			break;
		case MCV_CVP_PIPELINE:
			printf("----- TEST: CVP PIPELINE   -----\n");
			if(run_undistort_dfs_pipeline(undistort_config_l, undistort_config_r,
				dfs_config, imgL.data, imgR.data, disparity_rast, occlusion_rast))
			{
				fprintf(stderr, "ERROR in performing CVP undistort+DFS pipeline\n");
				return -1;
			}
			break;
	}

	// mcv_pc_disp_table_t pc_disp_table  = MCV_PC_DISP_TABLE_INITIALIZER;
	printf("linearizing disparity...\n");
	mcv_dfs_linearize_disparity(disparity_rast, w, h, disparity_lin);
	printf("done linearizing disparity, now occlusion...\n");
	mcv_dfs_linearize_occlusion(occlusion_rast, w, h, occlusion_lin);
	printf("done with occlusion...\n");

	Mat disp_rast_mat(h,w,CV_8UC1,disparity_rast);
	imwrite(DISPARITY_RAST_FILE, disp_rast_mat*4);
	Mat disp_mat(h,w,CV_8UC1,disparity_lin);
	imwrite(DISPARITY_FILE, disp_mat*4);
	Mat occ_mat(h,w,CV_8UC1,occlusion_rast);
	imwrite(OCCLUSION_FILE, occ_mat*128);

	// cleanup
	if(mcv_cvp_dfs_deinit(dfs_handle))
	{
		printf("Error: Failed to deinit DFS\n");
		return -1;
	}
	mcv_undistort_map_clean(&map_l);
	mcv_undistort_map_clean(&map_r);
	free(imgL_rect);
	free(imgR_rect);
	free(disparity_rast);
	free(occlusion_rast);
	free(disparity_lin);
	free(occlusion_lin);

	return 0;
}


/**
 * @brief Run the MCV tests on the input images, comparing hardware options
 * 
 * @return int 
 */
int main()
{
	// enable debug and timing printouts
	mcv_en_debug_prints(1);
	mcv_en_timing_prints(0);

	// initialize directories
	if(init_directories())
	{
		fprintf(stderr, "ERROR in %s\n", __FUNCTION__);
		return -1;
	}

	// run tests on each piece of hardware
	// run_test(MCV_CPU);
	// run_test(MCV_CVP);
	run_test(MCV_CVP_PIPELINE);

}


