/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file        mcv-test-feature-tracker.cpp
 *
 * @brief       Test code for the modalcv feature tracker
 *
 * @author      Thomas Patton (thomas.patton@modalai.com)
 * @date        2023
 */

#include <fcntl.h>
#include <getopt.h>
#include <modalcv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <iomanip>
#include <iostream>
#include <modalcv/io.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include <sstream>

#include "modalcv/feature_tracker.h"

#define OUTPUT_FOLDER "/data/libmodalcv/feature_tracker/"
#define N_DISPLAY 4096  // number of fp to draw
#define OLD_IMPL

int kclusters = 40; // Number of clusters

struct Point {
  double x;
  double y;
};

double distance(const Point& p1, const Point& p2);
std::vector<Point> kmeans(const std::vector<Point>& points, int k, double min_distance) ;


double distance(const Point& p1, const Point& p2) {
  return std::sqrt(std::pow(p1.x - p2.x, 2) + std::pow(p1.y - p2.y, 2));
}

// K-means clustering function
std::vector<Point> kmeans(const std::vector<Point>& points, int k, double min_distance) {
  // Initialize centroids randomly
  std::vector<Point> centroids(k);
  srand(time(0)); // Seed random number generator
  for (int i = 0; i < k; ++i) {
    int random_index = rand() % points.size();
    centroids[i] = points[random_index];
  }

  bool changed = true;
  while (changed) {
    changed = false;

    // Assign points to closest centroids
    std::vector<std::vector<Point>> clusters(k);
    for (const Point& point : points) {
      int closest_centroid = 0;
      double min_dist = distance(point, centroids[0]);
      for (int i = 1; i < k; ++i) {
        double curr_dist = distance(point, centroids[i]);
        if (curr_dist < min_dist) {
          min_dist = curr_dist;
          closest_centroid = i;
        }
      }
      // Check minimum distance constraint
      if (min_dist >= min_distance) {
        continue; // Point is too far from any centroid, skip for now
      }
      clusters[closest_centroid].push_back(point);
    }

    // Recompute centroids as the mean of points in each cluster
    std::vector<Point> new_centroids(k);
    for (int i = 0; i < k; ++i) {
      if (clusters[i].empty()) {
        // Handle empty clusters (re-initialize randomly)
        int random_index = rand() % points.size();
        new_centroids[i] = points[random_index];
      } else {
        double sum_x = 0, sum_y = 0;
        for (const Point& p : clusters[i]) {
          sum_x += p.x;
          sum_y += p.y;
        }
        new_centroids[i] = {sum_x / clusters[i].size(), sum_y / clusters[i].size()};
      }
    }

    // Check for convergence (no centroid movement)
    changed = false;
    for (int i = 0; i < k; ++i) {
      if (distance(centroids[i], new_centroids[i]) > 1e-6) {
        changed = true;
        break;
      }
    }

    centroids = new_centroids;
  }

  // Return only the centroids as the representative points
  return centroids;
}

static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
        fprintf(stderr, "ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec * 1000000000 + (int64_t)ts.tv_nsec;
}

static int _compare_features(const void* a, const void* b) {
    const mcv_cvp_feature_tracker_output_t* feature_a =
        (mcv_cvp_feature_tracker_output_t*)a;
    const mcv_cvp_feature_tracker_output_t* feature_b =
        (mcv_cvp_feature_tracker_output_t*)b;
    return (feature_b->age > feature_a->age);
}

/**
 * @brief Draws outputted features onto a color image
 *
 * @param img Image to draw on
 * @param output_features The output features to draw
 * @param img Number of outputted features
 */
static void draw_features(
    cv::Mat img,
	std::vector<Point> centroids)
{
    int radius, thickness;
    radius = 10;
    thickness = 3;
    for (int16_t i = 0; i < centroids.size(); i++) {
        cv::Scalar color(std::max(0, 255 - 5 * 3), 255, 0);
        cv::Rect rect(centroids[i].x - radius, centroids[i].y - radius, radius * 2,
                      radius * 2);
        cv::rectangle(img, rect, color, thickness, 1, 0);
    }
}

static void old_draw_features(
    cv::Mat img,
    mcv_cvp_feature_tracker_output_t output_features[MAX_N_FEATURES]) {

	for (int16_t i = 0; i < N_DISPLAY; i++) {
        mcv_cvp_feature_tracker_output_t feature = output_features[i];
        int age = feature.age;
        int radius, thickness;
        if (age > 15) {
            radius = 3;
            thickness = 3;
        } else {
            radius = 1;
            thickness = 3;
        }
        cv::Scalar color(std::max(0, 255 - 5 * age), 255, 0);
        cv::Rect rect(feature.x - radius, feature.y - radius, radius * 2,
                      radius * 2);
        cv::rectangle(img, rect, color, thickness, 1, 0);


		char str[32];
		int baseline = 0;
		float font_size = 0.5;
		sprintf(str, "%d(%d)", (int)feature.id, (int)feature.age);
		cv::Size text_size = cv::getTextSize(str, cv::FONT_HERSHEY_COMPLEX,
				font_size, font_size, &baseline);

		cv::putText(
				img, //target image
				str, //text
				cv::Point((feature.x - radius - text_size.width),
						feature.y - radius-text_size.height),
				cv::FONT_HERSHEY_COMPLEX, font_size,
				cv::Scalar(255, 255, 255), //font color
				font_size, cv::LINE_AA);

    }
}

int main(int argc, char *argv[]) {
    mcv_en_debug_prints(0);
    mcv_en_timing_prints(1);

    // vars to hold args
    int w, h;
    std::vector<std::string> paths;

    // parse input arguments and extract paths for feature extraction
    if (init_modalcv_example(argc, argv, paths, w, h)) {
        exit(1);
    }

    // configuration object for feature tracking
    mcv_cvp_feature_tracker_config_t feature_tracker_config;

    // which octave of the pyramid to perform optic flow on
    feature_tracker_config.opt_flow_octave = 1;

    // create a binary mask for features (NULL implies all 1's)
    feature_tracker_config.feature_mask = NULL;


    // set up feature point extraction
    mcv_cvp_pyr_fpx_config_t pyr_fpx_config = mcv_cvp_pyr_fpx_default_config();
    pyr_fpx_config.width = w;
    pyr_fpx_config.height = h;
    feature_tracker_config.pyr_fpx_config = pyr_fpx_config;

    // set up opt flow
    // w/h here are dictated by the octave we want to do flow on as defined
    // above in the feature_tracker_config
    feature_tracker_config.opt_flow_config.width =
        w / pow(2, feature_tracker_config.opt_flow_octave);
    feature_tracker_config.opt_flow_config.height =
        h / pow(2, feature_tracker_config.opt_flow_octave);
    feature_tracker_config.opt_flow_config.en_stats = 0;
    feature_tracker_config.opt_flow_config.input_ubwc =
        1;  // pyrfpx output ubwc

    // get a handle for our patch tracker
    mcv_cvp_feature_tracker_handle handle;
    handle = mcv_cvp_feature_tracker_init(feature_tracker_config);
    if (NULL == handle) {
        fprintf(stderr, "ERROR getting patch tracker handle\n");
        return -1;
    }

	cvp_set_pixel_search_radius(10);
	cvp_set_octave_search_radius(3);
	cvp_set_en_id_sort(0);

    //  iterate over all images in folder
    printf("Starting main loop...\n");
    for (size_t i = 0; i < paths.size(); i++) {
        // load the file
        cv::Mat current_img_gray, current_img_color;
        current_img_gray = imread(paths[i], cv::IMREAD_GRAYSCALE);
        if (current_img_gray.type() != CV_8UC1) {
            fprintf(stderr, "ERROR loading image\n");
            return -1;
        }
        int64_t start = _apps_time_monotonic_ns();
        cv::equalizeHist(current_img_gray, current_img_gray);
        cv::cvtColor(current_img_gray, current_img_color, cv::COLOR_GRAY2BGR);
        // cv::GaussianBlur(current_img_gray, current_img_gray,cv::Size(3, 3), 0);

        // cv::blur(current_img_gray, current_img_gray, cv::Size(5, 5));
 
        // some output things
        int16_t num_outputs;
        mcv_cvp_feature_tracker_output_t output[MAX_N_FEATURES];

        // run the patch tracker process on the current frame
        // mcv_cvp_feature_tracker_process(0, handle, current_img_gray.data,
        //                                 nullptr,
        //                                 30,
        //                                 &num_outputs, output);
        int64_t end = _apps_time_monotonic_ns();
        printf("Iter Time ::  %6.2fms  (%d)\n", (double)(end - start) / 1000000, num_outputs);
        printf("opt flow oct: %d\n", feature_tracker_config.opt_flow_octave);

        if (num_outputs > 0)
        {
		std::vector<Point> inpts(num_outputs);
		for (int y=0; y<num_outputs; y++)
		{
			inpts[y].x = output[y].x;
			inpts[y].y = output[y].y;
		}

		std::vector<Point> centroids = kmeans(inpts, kclusters, 200); // Minimum distance 50 pixels

		// draw features on the image
		// draw_features(current_img_color, centroids);

		// std::stringstream ss;
	  //       ss << std::setw(5) << std::setfill('0') << i;
	  //       std::string output_path = OUTPUT_FOLDER + ss.str() + ".png";
	  //       cv::imwrite(output_path, current_img_color);
        }

#ifdef OLD_IMPL
        old_draw_features(current_img_color, output);

		std::stringstream ss;
        ss << std::setw(5) << std::setfill('0') << i;
        std::string output_path = OUTPUT_FOLDER + ss.str() + ".png";
        printf("output: %s\n", output_path.c_str());
        cv::imwrite(output_path, current_img_color);
#endif

    }

    // cleanup
    printf("all done, deinitializing\n");
    mcv_cvp_feature_tracker_deinit(handle);
    return 0;
}
