/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file       mcv-test-opt-flow-fpx.cpp
 *
 * @brief      Tests the Optic Flow and PYR Feature Extraction (FPX) process
 *
 * @author     Kyle Tyni (kyle.tyni@modalai.com)
 * @date       2024
 */


#include <fastcv.h>
#include <fcntl.h>
#include <getopt.h>
#include <modalcv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>   // for mkdir
#include <sys/types.h>  // for mode_t in mkdir
#include <time.h>
#include <unistd.h>
#include <cmath>

#include <experimental/filesystem>
#include <iomanip>
#include <iostream>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/opencv.hpp>
#include <unordered_map>
#include <sstream>
#include <random>
#include <modalcv/io.hpp>
#include "modalcv/pyr.h"

namespace fs = std::experimental::filesystem;

#define OUTPUT_FOLDER		   	"/data/libmodalcv/fpx-flow/"
#define NORMALIZED_GAIN_FOLDER 	"/data/libmodalcv/fpx-flow/norm_gain/"
#define OPT_FLOW_RESULT_FOLDER 	"/data/libmodalcv/fpx-flow/res/"

const std::string OCTAVE_FOLDERS[] = {
    std::string(OUTPUT_FOLDER) + "octave_0/",
    std::string(OUTPUT_FOLDER) + "octave_1/",
    std::string(OUTPUT_FOLDER) + "octave_2/"
};

const bool do_gain_normalization = false;

const int opt_flow_octave = 1;
const bool draw_cv_vectors = false;

int create_output_dirs();
int normalize_gains(std::vector<std::string> paths);
int do_pyramid_extraction(mcv_cvp_pyr_handle& handle, 
						  mcv_cvp_pyr_t& pyr, 
						  mcv_cvp_pyr_config_t& pyr_config,
						  std::vector<std::string> paths);

int do_optic_flow(mcv_cvp_opt_flow_handle& handle, mcv_cvp_opt_flow_output_t& output);
double calculate_brightness(const cv::Mat& frame);
void get_avg_vector(std::vector<mcv_cvp_motion_vector_t> vectors, mcv_cvp_motion_vector_t& res,
                    float& stddev_x, float& stddev_y);


/**
 * @brief Gets the system time
 * 
 * @return int64_t The system time
 */
static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
        fprintf(stderr, "ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec * 1000000000 + (int64_t)ts.tv_nsec;
}

/**
 * @brief Reads in images from dir and extracts data
 */
static int _get_images_from_dir(const std::string input_folder,
                                std::vector<std::string>& matching_paths,
                                int& w, int& h) {
    const std::string extension = ".png";
    try {
        for (const auto& entry : fs::directory_iterator(input_folder)) {
            // check if entry is regular file and matches extension
            if (fs::is_regular_file(entry) &&
                entry.path().extension() == extension) {
                matching_paths.push_back(entry.path().filename().string());
            }
        }
        // sort the filenames
        std::sort(matching_paths.begin(), matching_paths.end());
    } catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << std::endl;
        return 1;
    }
    cv::Mat tmp_img =
        cv::imread(input_folder + matching_paths[0], cv::IMREAD_GRAYSCALE);
    w = tmp_img.cols;
    h = tmp_img.rows;
    return 1;
}

/**
 * @brief A method for drawing the output of our optic flow method
 */
static void draw_opt_flow(
    cv::Mat                                           img,
    mcv_cvp_motion_vector_t*                          motion_vector,
    std::vector<int>                                  outlier_i,
    std::unordered_map<int, cv::Point2f>              outlier_mvs,
    __attribute__((unused)) mcv_cvp_opt_flow_stats_t* stats,
    cv::Mat                                           cv_flow,
    int                                               w,
    int                                               h) {
    int n_rows = h / 8;
    int n_cols = w / 8; // 8x8 block

    printf("num outliers: %d\n", outlier_mvs.size());

    int scale_factor = std::pow(2, opt_flow_octave);

    // draw the generated motion vectors
    for (int row_i = 0; row_i < n_rows; row_i++) {
        for (int col_i = 0; col_i < n_cols; col_i++) {

            int arrow_start_x = (col_i * 8 + 4);
            int arrow_start_y = (row_i * 8 + 4);
            int coord         = x_y_to_flat(col_i, row_i, n_cols, n_rows, 1);
            
            // skip if coord has and index present in outlier_i

            int arrow_end_x   = (arrow_start_x - motion_vector[coord].mv_x);
            int arrow_end_y   = (arrow_start_y - motion_vector[coord].mv_y);

            cv::Scalar mv_color;
            if (std::binary_search(outlier_i.begin(), outlier_i.end(), coord)) {
                mv_color = cv::Scalar(0, 0, 255);

                cv::line(img, cv::Point2f(arrow_start_x * scale_factor, arrow_start_y * scale_factor),
                    cv::Point2f((arrow_start_x - static_cast<int>(std::round(outlier_mvs[coord].x))) * scale_factor, 
                                (arrow_start_y - static_cast<int>(std::round(outlier_mvs[coord].y))) * scale_factor),
                    cv::Scalar(0, 255, 255), 1, 1);

            } else {
                mv_color = cv::Scalar(0, 255, 0);
            }

            cv::line(
                img,
                cv::Point2f(arrow_start_x * scale_factor, arrow_start_y * scale_factor),
                cv::Point2f(arrow_end_x * scale_factor, arrow_end_y * scale_factor),
                mv_color, 1, 1);

            cv::circle(img, cv::Point(arrow_start_x * scale_factor, arrow_start_y * scale_factor), 2,
                       mv_color, -1);
        }
    }

    if (draw_cv_vectors) {

    int block_size = 8;
        for (int y = 0; y < h; y += block_size) {
            for (int x = 0; x < w; x += block_size) {
                cv::Point2f avg_flow(0.0, 0.0);
                int count = 0;

                for (int j = 0; j < block_size && y + j < h; ++j) {
                    for (int i = 0; i < block_size && x + i < w; ++i) {
                        cv::Point2f flow_at_point = cv_flow.at<cv::Point2f>(y + j, x + i);
                        avg_flow += flow_at_point;
                        count++;
                    }
                }

                avg_flow.x /= count;
                avg_flow.y /= count;

                // Draw the motion vector as an arrow
                cv::Point start_point(x + block_size / 2, y + block_size / 2);
                cv::Point end_point = start_point + cv::Point(static_cast<int>(avg_flow.x), static_cast<int>(avg_flow.y));
                // cv::line(img, start_point * scale_factor, end_point * scale_factor, cv::Scalar(255, 255, 0), 1, 1);
            }
        }
    }
}

int main(int argc, char* argv[]) {
	mcv_en_debug_prints(0);
	mcv_en_timing_prints(0);
    
	int w, h;
    std::vector<std::string> paths;

    // args checks, makes sure path is good
    if (init_modalcv_example(argc, argv, paths, w, h)) {
        fprintf(stderr, "ERROR, failed to initialize example\n");
        return -1;
    }

    // creates directories for image outputs
    if (create_output_dirs() != 0) {
		fprintf(stderr, "ERROR in create_output_dirs()\n");
		return -1;
	}
    
    /********* NORMALIZE GAIN ACROSS VIDEO *********/

    if (do_gain_normalization) {
        normalize_gains(paths);
    }

	/********* PYRAMID EXTRACTION *********/

    // init pyramid extractor config
    mcv_cvp_pyr_config_t pyr_config;
	pyr_config.width               = w;
	pyr_config.height              = h;
	pyr_config.n_octaves           = 3;
	pyr_config.n_scales_per_octave = 1;

	// gets a handle for pyramid downscaler process
	mcv_cvp_pyr_handle pyr_handle = mcv_cvp_pyr_init(pyr_config);
	if(NULL == pyr_handle) {
		fprintf(stderr, "ERROR creating pyramid handle\n");
		return -1;
	}

	// creates output structure
	mcv_cvp_pyr_t pyr;
	if (mcv_cvp_create_pyr(pyr_handle, &pyr)) {
		fprintf(stderr, "ERROR creating PYR output object\n");
		return -1;
	}

    // create downscaled images
    if (do_pyramid_extraction(pyr_handle, pyr, pyr_config, paths) != 0) {
		fprintf(stderr, "ERROR in do_pyramid_extraction()\n");
		return -1;
	}

    // cleanup for pyramid extraction
    if(mcv_cvp_destroy_pyr(pyr_handle, &pyr)) {
		fprintf(stderr, "ERROR destorying pyramid object\n");
		return -1;
	}
	if(mcv_cvp_pyr_deinit(pyr_handle)) {
		fprintf(stderr, "ERROR deinitializing pyramid\n");
		return -1;
	}


	/********* OPTICAL FLOW *********/

	// init optic flow config
    mcv_cvp_opt_flow_config_t opt_flow_config;
    opt_flow_config.width      = w >> opt_flow_octave; // this is because the image is second octave
    opt_flow_config.height     = h >> opt_flow_octave;
    opt_flow_config.input_ubwc = 0;
    opt_flow_config.en_stats   = 1;

	// gets a handle for optical flow
    mcv_cvp_opt_flow_handle opt_flow_handle = mcv_cvp_opt_flow_init(opt_flow_config);
    if (NULL == opt_flow_handle) {
        fprintf(stderr, "ERROR getting opt flow handle\n");
        return -1;
    }

    // create an optic flow output object
    mcv_cvp_opt_flow_output_t opt_flow_out;
    mcv_cvp_create_opt_flow_output(opt_flow_handle, &opt_flow_out);

	// run optic flow on the octave 2 images
	do_optic_flow(opt_flow_handle, opt_flow_out);

	// optic flow cleanup
    mcv_cvp_destroy_opt_flow_output(opt_flow_handle, &opt_flow_out);
    mcv_cvp_opt_flow_deinit(opt_flow_handle);

    // do fpx

    
    return 0;
}

double calculate_brightness(const cv::Mat& frame) {
    return cv::mean(frame)[0];
}

int normalize_gains(std::vector<std::string> paths) {
    std::vector<cv::Mat> frames;
    std::vector<double> brightness_values;
    double target_brightness = 0.0;

    // Load Frames
    for (size_t i = 0; i < paths.size(); i++) {
        cv::Mat frame = cv::imread(paths[i], cv::IMREAD_GRAYSCALE);
        if (!frame.empty()) {
            frames.push_back(frame);
            double brightness = calculate_brightness(frame);
            brightness_values.push_back(brightness);
            target_brightness += brightness;
        }
    }

    if (frames.empty()) {
        std::cout << "Error: No frames found in the folder.\n";
        return -1;
    }

    target_brightness /= frames.size();

    // Normalize Frames
    for (size_t i = 0; i < frames.size(); ++i) {
        cv::Mat normalized_frame;
        double current_brightness = brightness_values[i];
        double brightness_ratio = target_brightness / current_brightness;
        frames[i].convertTo(normalized_frame, -1, brightness_ratio, 0);

        // Save the normalized frame as JPEG

        // draw optic flow vectors and write to disk
        std::stringstream ss;
        ss << std::setw(5) << std::setfill('0') << i;
        std::string output_path = NORMALIZED_GAIN_FOLDER + ss.str() + ".png";
        // printf("%s\n", output_path.c_str());
        cv::imwrite(output_path, normalized_frame);
    }

    return 0;
}

int create_output_dirs() {
	if (std::system("mkdir -p /data/libmodalcv/fpx-flow")) {
        std::cout << "Error creating directories" << std::endl;
        return -1;
    }
    if (std::system("rm -rf /data/libmodalcv/pyr/fpx-flow/*")) {
        fprintf(
            stderr,
            "Warning, failed to delete old files from %s\n",
            OUTPUT_FOLDER);
        return -1;
    }

    if (std::system("mkdir -p /data/libmodalcv/fpx-flow/norm_gain")) {
        std::cout << "Error creating directories" << std::endl;
        return -1;
    }
    if (std::system("rm -rf /data/libmodalcv/pyr/fpx-flow/norm_gain/*")) {
        fprintf(
            stderr,
            "Warning, failed to delete old files from %s\n",
            OUTPUT_FOLDER);
        return -1;
    }

	if (std::system("mkdir -p /data/libmodalcv/fpx-flow/res")) {
        std::cout << "Error creating directories" << std::endl;
        return -1;
    }
    if (std::system("rm -rf /data/libmodalcv/pyr/fpx-flow/res/*")) {
        fprintf(
            stderr,
            "Warning, failed to delete old files from %s\n",
            OUTPUT_FOLDER);
        return -1;
    }

    for (int i = 0; i < 3; ++i) {
        // Create directories
		std::string mkdirCommand = "mkdir -p " + OCTAVE_FOLDERS[i];
        if (std::system(mkdirCommand.c_str())) {
            std::cout << "Error creating directory: " << OCTAVE_FOLDERS[i] << std::endl;
            return -1;
        }
        
        // Remove old files
		std::string rmCommand = "rm -rf " + OCTAVE_FOLDERS[i] + "*";
        if (std::system(rmCommand.c_str())) {
            std::cerr << "Warning, failed to delete old files from " << OCTAVE_FOLDERS[i] << std::endl;
            return -1;
        }
    }

    return 0;
}

int do_pyramid_extraction(mcv_cvp_pyr_handle& handle, 
						   mcv_cvp_pyr_t& pyr, 
						   mcv_cvp_pyr_config_t& pyr_config,
						   std::vector<std::string> p) {

    printf("Performing pyramid extraction\n");

    int w, h;
	std::vector<std::string> paths;

    if (do_gain_normalization) {
        _get_images_from_dir(NORMALIZED_GAIN_FOLDER, paths, w, h);
        for (size_t i = 0; i < paths.size(); i++) {
            paths[i] = NORMALIZED_GAIN_FOLDER + paths[i];
            std::cout << paths[i] << std::endl;
        }
    } else {
        paths = p;
    }

	for (size_t i = 0; i < paths.size(); i++) {
		// int64_t start = _apps_time_monotonic_ns();
        cv::Mat img_mat = cv::imread(paths[i], cv::IMREAD_GRAYSCALE);

		if (img_mat.type()!= CV_8UC1) {
            fprintf(stderr, "ERROR loading %s\n", paths[i].c_str());
            return -1;
        }

        int64_t start = _apps_time_monotonic_ns();
        cv::equalizeHist(img_mat, img_mat);
        cv::Mat filtered_img;
        // cv::bilateralFilter(img_mat, filtered_img, 9, 75, 75);
        // img_mat = filtered_img.clone();

        cv::Mat blurred;
        cv::GaussianBlur(img_mat, img_mat, cv::Size(9, 9), 1.0);
        // cv::addWeighted(img_mat, 1.5, blurred, -0.5, 0, img_mat);

        int64_t end = _apps_time_monotonic_ns();
        printf("Pre Proc Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

		// do the pyramid extraction on an individual image
		if(mcv_cvp_pyr_process(handle, img_mat.data, &pyr)) {
			fprintf(stderr, "ERROR executing pyramid process\n");
			return -1;
		}

		// int64_t end = _apps_time_monotonic_ns();
		// printf("Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

		// save to disk
		for (int j = 0; j < (int)pyr_config.n_octaves; j++) {
        	std::stringstream ss;
			ss << std::setw(5) << std::setfill('0') << i;
			std::string output_path = OCTAVE_FOLDERS[j] + ss.str() + ".png";
			// printf("%s\n", output_path.c_str());

			cv::Mat curr_pyr(pyr.height[j], pyr.width[j], CV_8UC1, 
							 pyr.image_pyramids[j]);

			cv::imwrite(output_path, curr_pyr);
		}
	}

	return 0;
}

cv::Mat estimateHomography(const std::vector<cv::Point2f>& points1, const std::vector<cv::Point2f>& points2);

int do_optic_flow(mcv_cvp_opt_flow_handle& handle, mcv_cvp_opt_flow_output_t& output) {
	int w, h;
    printf("starting opt flow\n");
    // load in octave 0 img paths
	std::vector<std::string> paths_oct0;
	_get_images_from_dir(OCTAVE_FOLDERS[0], paths_oct0, w, h);
	for (size_t i = 0; i < paths_oct0.size(); i++) {
        paths_oct0[i] = OCTAVE_FOLDERS[0] + paths_oct0[i];
    }

    // load in octave 2 img paths (level that opt flow runs on)
	std::vector<std::string> target_paths;
	_get_images_from_dir(OCTAVE_FOLDERS[opt_flow_octave], target_paths, w, h);
	for (size_t i = 0; i < target_paths.size(); i++) {
        target_paths[i] = OCTAVE_FOLDERS[opt_flow_octave] + target_paths[i];
    }

    // run opt flow on each image on octave 2
	int initialized = 0;
    for (size_t i = 1; i < target_paths.size(); i++) { // note opt flow needs to start from 1
	
        cv::Mat curr_img_gray, curr_img_color;
        std::cout << target_paths[i] << std::endl;
        curr_img_gray = cv::imread(target_paths[i], cv::IMREAD_GRAYSCALE);
		curr_img_color = cv::imread(paths_oct0[i], cv::IMREAD_GRAYSCALE);
        cv::cvtColor(curr_img_color, curr_img_color, cv::COLOR_GRAY2BGR);
        int64_t start = _apps_time_monotonic_ns();

        // used to compare flow calculted with opencv
        cv::Mat cv_flow;


        cv::Mat prev_img_gray;
        prev_img_gray = cv::imread(target_paths[i - 1], cv::IMREAD_GRAYSCALE);

        if (draw_cv_vectors) {
            cv::calcOpticalFlowFarneback(prev_img_gray, curr_img_gray, cv_flow, 0.5, 3, 15, 3, 5, 1.2, 0);
            printf("cv optic flow complete\n");
        }

        if (initialized) {
            // run opt flow using current img and img data in memory
            mcv_cvp_opt_flow_process_from_mem(
                handle, curr_img_gray.data, &output);

        } else {

            // run opt flow using prev and current img
            mcv_cvp_opt_flow_process_from_img(
                handle, prev_img_gray.data, curr_img_gray.data, &output);

            // image is in memory now
            initialized = 1;
        }
        int64_t end = _apps_time_monotonic_ns();
        printf("Opt Flow Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

        // do ransac to identify motion vector outliers
        const float inlier_threshold = 10.f; // could be avg mv magnitude
        const float success_probabilty = 0.95f;
        const float outlier_ratio = 0.45f;
        const int min_point_for_model = 4;

        // number of iterations to find ideal model
        int N = static_cast<int>(std::log(1 - success_probabilty) / std::log(1 - std::pow(1 - outlier_ratio, min_point_for_model)));

        cv::Mat best_homography;
        int max_inliers = 0;

        std::vector<cv::Point2f> coords;
        std::vector<cv::Point2f> coords_prime;

        int n_rows = h / 8;
        int n_cols = w / 8; // 8x8 block

        // draw the generated motion vectors
        for (int row_i = 0; row_i < n_rows; row_i++) {
            for (int col_i = 0; col_i < n_cols; col_i++) {

                int x           = (col_i * 8 + 4);
                int y           = (row_i * 8 + 4);
                int coord_i     = x_y_to_flat(col_i, row_i, n_cols, n_rows, 1);
                int dx          = (x + output.mvs[coord_i].mv_x);
                int dy          = (y + output.mvs[coord_i].mv_y);

                coords.push_back(cv::Point2f(x, y));
                coords_prime.push_back(cv::Point2f(dx, dy));
            }
        }

        std::random_device rd;
        std::mt19937 rng(rd());

        std::vector<int> outlier_indices_of_best_homography;
        std::unordered_map<int, cv::Point2f> best_replacement_mvs;
       /*

        for (int j = 0; j < N; ++j) {
            std::vector<int> indices(coords.size());
            std::iota(indices.begin(), indices.end(), 0);
            std::shuffle(indices.begin(), indices.end(), rng);

            std::vector<cv::Point2f> sample1(4);
            std::vector<cv::Point2f> sample2(4);

            for (int i = 0; i < 4; ++i) {
                sample1[i] = coords[indices[i]];
                sample2[i] = coords_prime[indices[i]];
            }

            // Step 2: Estimate homography from the sample
            cv::Mat H = estimateHomography(sample1, sample2);

            // Step 3: Score the model by counting inliers
            int inliers_count = 0;

            std::vector<int> outlier_indices;
            std::unordered_map<int, cv::Point2f> outlier_replacement_mvs;

            for (size_t i = 0; i < coords.size(); ++i) {
                // Project points1[i] using H
                std::vector<cv::Point2f> projected_point;
                cv::perspectiveTransform(std::vector<cv::Point2f>{coords[i]}, projected_point, H);

                // Calculate the distance between projected point and points2[i]
                float dist = cv::norm(coords_prime[i] - projected_point[0]);

                if (dist < inlier_threshold) {
                    ++inliers_count;
                    // printf("inlear found, dist: %.3f, org: (%.2f,%.2f) prime:(%.2f,%.2f)\n", dist,
                    //         coords[i].x, coords[i].y, 
                    //         projected_point[0].x, projected_point[0].y);
                } else {
                    outlier_indices.push_back(i);
                    cv::Point2f motion_vector = projected_point[0] - coords[i];
                    outlier_replacement_mvs[i] = motion_vector;
                }
            }

            // Step 4: Update the best model if this one has more inliers
            if (inliers_count > max_inliers) {
                max_inliers = inliers_count;
                best_homography = H;
                outlier_indices_of_best_homography = outlier_indices;
                best_replacement_mvs = outlier_replacement_mvs;
            }
        }

        if (best_replacement_mvs.size() > 0) {
            std::cout << "Max Inliers: " << max_inliers << ", Max Outliers: " << outlier_indices_of_best_homography.size() << std::endl;
            printf("replacement mvs: %d\n", best_replacement_mvs.size());

            for (const auto& pair : best_replacement_mvs) {
                int c = pair.first;
                // printf("coord_i: %d, xy:(%.2f, %.2f) mv: (%.2f, %.2f)\n", c, 
                // coords[c].x, coords[c].y, pair.second.x, pair.second.y);
            }
        }
        // std::cout << "Best Homography:\n" << best_homography << std::endl;
       */

        // int64_t end = _apps_time_monotonic_ns();
        // printf("Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

        // draw optic flow vectors and write to disk
        std::stringstream ss;
        ss << std::setw(5) << std::setfill('0') << i;
        std::string output_path = OPT_FLOW_RESULT_FOLDER + ss.str() + ".png";

        // outlier_indices_of_best_homography.clear();
        // best_replacement_mvs.clear();
		
        draw_opt_flow(
            curr_img_color,
            output.mvs,
            outlier_indices_of_best_homography,
            best_replacement_mvs,
            output.stats,
            cv_flow,
            curr_img_gray.cols,
            curr_img_gray.rows);
        cv::imwrite(output_path, curr_img_color);
    }


    return 0;
}





cv::Mat estimateHomography(const std::vector<cv::Point2f>& points1, const std::vector<cv::Point2f>& points2) {
    CV_Assert(points1.size() == 4 && points2.size() == 4);

    cv::Mat A(8, 9, CV_32F);

    for (int i = 0; i < 4; i++) {
        float x = points1[i].x;
        float y = points1[i].y;
        float x_prime = points2[i].x;
        float y_prime = points2[i].y;

        A.at<float>(2 * i, 0) = x;
        A.at<float>(2 * i, 1) = y;
        A.at<float>(2 * i, 2) = 1.0;
        A.at<float>(2 * i, 3) = 0.0;
        A.at<float>(2 * i, 4) = 0.0;
        A.at<float>(2 * i, 5) = 0.0;
        A.at<float>(2 * i, 6) = -x * x_prime;
        A.at<float>(2 * i, 7) = -y * x_prime;
        A.at<float>(2 * i, 8) = -x_prime;

        A.at<float>(2 * i + 1, 0) = 0.0;
        A.at<float>(2 * i + 1, 1) = 0.0;
        A.at<float>(2 * i + 1, 2) = 0.0;
        A.at<float>(2 * i + 1, 3) = x;
        A.at<float>(2 * i + 1, 4) = y;
        A.at<float>(2 * i + 1, 5) = 1.0;
        A.at<float>(2 * i + 1, 6) = -x * y_prime;
        A.at<float>(2 * i + 1, 7) = -y * y_prime;
        A.at<float>(2 * i + 1, 8) = -y_prime;
    }

    // Solve the system using SVD
    cv::Mat h;
    cv::SVD::solveZ(A, h);

    cv::Mat H = (cv::Mat_<float>(3, 3) << 
        h.at<float>(0), h.at<float>(1), h.at<float>(2),
        h.at<float>(3), h.at<float>(4), h.at<float>(5),
        h.at<float>(6), h.at<float>(7), h.at<float>(8));

    return H;
}