/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

/**
 * @file       mcv-test-fpx.cpp <examples/mcv-test-undistort>
 *
 * @brief      Tests the Feature Extraction (FPX) process
 * 
 *             This test is intended to be run with a path to some voxl-logger
 *             dataset as an argument. 
 * 
 * @author     Thomas Patton (thomas.patton@modalai.com)
 * @date       2023
 */

#include <stdio.h>
#include <sstream>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/stat.h>	// for mkdir
#include <sys/types.h>	// for mode_t in mkdir
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <iomanip>
#include <iostream>
#include <modalcv.h>
#include <modalcv/io.hpp>

#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>

#define OUTPUT_FOLDER	"/data/libmodalcv/fpx/"

void draw_color_plus_sign(cv::Mat &img, int x, int y, int score, int size = 3, int thickness = 1);

/**
 * @brief Helper fn for getting the current clock time in nanoseconds
 *
 * @return Time in ns
 */
static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if(clock_gettime(CLOCK_MONOTONIC, &ts)){
        fprintf(stderr,"ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec*1000000000 + (int64_t)ts.tv_nsec;
}

int main(int argc, char* argv[]) {
    mcv_en_debug_prints(1);
    mcv_en_timing_prints(1);

    int w, h;
    std::vector<std::string> paths;

    // args checks, makes sure path is good
    if (init_modalcv_example(argc, argv, paths, w, h)) {
        fprintf(stderr, "ERROR, failed to initialize example\n");
        return -1;
    }

    // clean output folder
    if (std::system("mkdir -p /data/libmodalcv/fpx")) {
        std::cout << "Error creating directories" << std::endl;
        return -1;
    }
    if (std::system("rm -rf /data/libmodalcv/fpx/*")) {
        fprintf(
            stderr,
            "Warning, failed to delete old files from %s\n",
            OUTPUT_FOLDER);
        return -1;
    }

    // initialize feature point extractor
    mcv_cvp_fpx_config_t fpx_config;
    fpx_config.width           = w;
    fpx_config.height          = h;
    // fpx_config.mode            = MCV_CVP_FPX_PEAK_8x8;
    fpx_config.mode            = MCV_CVP_FPX_ZONE;
    fpx_config.nms_mode        = MCV_CVP_FPX_5_TAP_NMS;
    fpx_config.score_threshold = 983;                    // number as a percentage of 2^16-1
    fpx_config.robustness      = 10;                     // 0-127, default 10

    // get a handle for our FPX process
    mcv_cvp_fpx_handle fpx_handle = mcv_cvp_fpx_init(fpx_config);
    if(fpx_handle == NULL) {
        fprintf(stderr, "ERROR getting FPX handle\n");
        return -1;
    }

    // create output object
    mcv_cvp_fpx_t fpx;
    if(mcv_cvp_create_fpx(fpx_handle, &fpx)) {
        fprintf(stderr, "ERROR creating FPX output object\n");
        return -1;
    }

    // iterate over each image and extract features
    printf("starting main loop...\n");
    for (size_t i = 0; i < paths.size(); i++) {
        int64_t start = _apps_time_monotonic_ns();
        
        // load image from path, make a color capable img for visuals
        cv::Mat img_mat_color;
        cv::Mat img_mat = cv::imread(paths[i], cv::IMREAD_GRAYSCALE);
        cv::cvtColor(img_mat, img_mat_color, cv::COLOR_GRAY2BGR);

        if (img_mat.type()!= CV_8UC1) {
            fprintf(stderr, "ERROR loading %s\n", paths[i].c_str());
            return -1;
        }

        // do the feature point extraction
        if (mcv_cvp_fpx_process(fpx_handle, img_mat.data, &fpx)) {
            fprintf(stderr, "ERROR completing FPX process\n");
            return -1;
        }
        int64_t end = _apps_time_monotonic_ns();
        printf("Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

        // draw overlay and write to disk
        std::stringstream ss;
        ss << std::setw(5) << std::setfill('0') << i;
        std::string output_path = OUTPUT_FOLDER + ss.str() + ".png";
        printf("%s\n", output_path.c_str());
        for(int j = 0; j < (int) fpx.n_points; j++){
            int x = fpx.features[j].x;
            int y = fpx.features[j].y;
            draw_color_plus_sign(img_mat_color, x, y, fpx.features[j].score);

            // draw max score at bottom left
            int fontFace = cv::FONT_HERSHEY_SIMPLEX;
            cv::Scalar color(255, 255, 0); // Green color in BGR
            int thickness = 1; // Thickness of the text

            int baseline = 0;
            std::string text = std::to_string(fpx.max_score);
            cv::Size textSize = cv::getTextSize(text, fontFace, 0.5, thickness, &baseline);
            cv::Point position(10, h - 10);

            // Put the text on the image
            cv::putText(img_mat_color, text, position, fontFace, 0.5, color, thickness, cv::LINE_AA);

            // mcv_overlay_draw_plus(img_mat.data, w, h, x, y); // this function doesn't acount for 3 channel img

        }

        cv::imwrite(output_path, img_mat_color);
    }

    // cleanup
    mcv_cvp_fpx_deinit(fpx_handle);

    // printf("setting permissions for output files\n");
    // ret = system("chmod 777 -R /data/libmodalcv/");
    // if(ret){
    //     fprintf(stderr, "Warning, failed to set permissions for output files\n");
    // }

    printf("DONE\n");
    return 0;
}

void draw_color_plus_sign(cv::Mat &img, int x, int y, int score, int size, int thickness) {

    // could be used for a color gradient based on the score of points
    cv::Scalar innerColor = cv::Scalar(255, 255, 255); // White color (B, G, R)
    cv::Scalar borderColor = cv::Scalar(0, 0, 0);     // Black color (B, G, R)
    
    int borderThickness = thickness + 1; // Border thickness slightly larger than the inner part
    
    // Draw the black border
    cv::line(img, cv::Point(x - size, y), cv::Point(x + size, y), borderColor, borderThickness); // Horizontal line
    cv::line(img, cv::Point(x, y - size), cv::Point(x, y + size), borderColor, borderThickness); // Vertical line
    
    // Draw the white inner part
    cv::line(img, cv::Point(x - size, y), cv::Point(x + size, y), innerColor, thickness); // Horizontal line
    cv::line(img, cv::Point(x, y - size), cv::Point(x, y + size), innerColor, thickness); // Vertical line
}
