/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file       mcv-test-ft.cpp
 *
 * @brief      Tests the feature tracking process using fastcv functions
 *
 * @author     Kyle Tyni (kyle.tyni@modalai.com)
 * @date       2024
 */


#include <iostream>

#include <modal_pipe.h>
#include <modalcv.h>

#include <fastcv.h>

#include <opencv2/opencv.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/core/ocl.hpp>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <vector>
#include <string>
#include <algorithm>
#include <experimental/filesystem>

namespace fs = std::experimental::filesystem;


// Function to retrieve and sort PNG files from a directory using experimental filesystem
std::vector<fs::path> getSortedPNGFiles(const std::string& folderPath) {
    std::vector<fs::path> pngFiles;

    // Iterate through the directory
    for (const auto& entry : fs::directory_iterator(folderPath)) {
        if (fs::is_regular_file(entry.status())) {
            // Check for .png extension (case-insensitive)
            if (entry.path().extension() == ".png" || entry.path().extension() == ".PNG") {
                pngFiles.emplace_back(entry.path());
            }
        }
    }

    // Sort the files by filename
    std::sort(pngFiles.begin(), pngFiles.end());

    return pngFiles;
}

struct Point {
    float x;
    float y;
};

// Struct to represent a feature with ID and age
struct Feature {
    int id;             // Unique identifier
    Point point;        // Current position
    int age;            // Number of frames tracked
    float score = 0.f;  // score of the feature point
};

// Function to find top N oldest features
std::vector<Feature> getTopNOldestFeatures(const std::vector<Feature>& features, int N) {
    // Create a copy of features to sort
    std::vector<Feature> sortedFeatures = features;

    // Sort features based on age in descending order
    std::sort(sortedFeatures.begin(), sortedFeatures.end(), 
        [](const Feature& a, const Feature& b) -> bool {
            if (a.age != b.age) {
                return a.age > b.age; // Sort by age first (descending)
            } else {
                return a.score > b.score; // If ages are equal, sort by score (descending)
            }
        }
    );

    // Select top N features
    if (sortedFeatures.size() > N) {
        sortedFeatures.resize(N);
    }

    return sortedFeatures;
}

std::vector<bool> isLowGradient(const cv::Mat& img, std::vector<Feature> features, double threshold = 40.0, int blockSize = 3) {
    std::vector<bool> results;
    results.reserve(features.size());

    // Calculate gradients using Sobel operator
    cv::Mat gx, gy;
    cv::Sobel(img, gx, CV_64F, 1, 0, 3); // Gradient in x-direction
    cv::Sobel(img, gy, CV_64F, 0, 1, 3); // Gradient in y-direction

    for (const auto& feat : features) {
        int x = std::round(feat.point.x);
        int y = std::round(feat.point.y);
        
        // Ensure the point is within the bounds of the image and blockSize is valid
        int halfBlockSize = blockSize / 2;
        if (x < halfBlockSize || y < halfBlockSize || 
            x >= img.cols - halfBlockSize || y >= img.rows - halfBlockSize) {
            results.push_back(true);
            continue;; // If out of bounds, consider it low gradient to avoid tracking
        }

        // Calculate sum of squared gradients in the local block
        double sumSquaredGradients = 0.0;
        for (int j = -halfBlockSize; j <= halfBlockSize; ++j) {
            for (int i = -halfBlockSize; i <= halfBlockSize; ++i) {
                double dx = gx.at<double>(y + j, x + i);
                double dy = gy.at<double>(y + j, x + i);
                sumSquaredGradients += dx * dx + dy * dy;
            }
        }

        // If the sum of squared gradients is below the threshold, it is a low gradient point
        results.push_back(sumSquaredGradients < threshold);
    }

    return results;
}


static uint64_t _apps_time_monotonic_ns()
{
    struct timespec ts;
    if(clock_gettime(CLOCK_MONOTONIC, &ts)){
        fprintf(stderr,"ERROR calling clock_gettime\n");
        return -1;
    }
    return (uint64_t)ts.tv_sec*1000000000 + (uint64_t)ts.tv_nsec;
}

// Function to calculate Euclidean distance between two points
float calculateDistance(const cv::Point2f& p1, const cv::Point2f& p2) {
    float dx = p1.x - p2.x;
    float dy = p1.y - p2.y;
    return std::sqrt(dx * dx + dy * dy);
}


unsigned char* getAlignedData(const cv::Mat& image, bool& needToFree) {
    if (reinterpret_cast<uintptr_t>(image.data) % 16 == 0) {
        needToFree = false;
        return image.data;
    } else {
        size_t totalSize = image.total() * image.elemSize();
        unsigned char* alignedData = nullptr;
        if (posix_memalign(reinterpret_cast<void**>(&alignedData), 16, totalSize) != 0) {
            needToFree = false;
            return nullptr; // Allocation failed
        }
        memcpy(alignedData, image.data, totalSize);
        needToFree = true;
        return alignedData;
    }
}

// Function to calculate NCC between two image patches
double calculateNCC(const cv::Mat& patch1, const cv::Mat& patch2) {
    cv::Mat patch1D, patch2D;
    patch1.convertTo(patch1D, CV_32F);
    patch2.convertTo(patch2D, CV_32F);

    // Subtract mean from patches
    patch1D -= cv::mean(patch1D);
    patch2D -= cv::mean(patch2D);

    // Calculate NCC
    double numerator = patch1D.dot(patch2D);
    double denominator = std::sqrt(patch1D.dot(patch1D) * patch2D.dot(patch2D));
    
    if (denominator == 0) return 0.0; // Avoid division by zero
    return numerator / denominator;
}


int main(int argc, char* argv[]) {
    // Check for proper usage
    if (argc != 2) {
        std::cerr << "Usage: " << argv[0] << " <path_to_png_folder>" << std::endl;
        return -1;
    }

    std::string folderPath = argv[1];

    // Get sorted list of PNG files
    std::vector<fs::path> pngFiles = getSortedPNGFiles(folderPath);

    if (pngFiles.empty()) {
        std::cerr << "No PNG files found in the specified folder." << std::endl;
        return -1;
    }

    // Variables to hold previous and current images
    cv::Mat prevGray, currGray;

    // Feature tracking variables
    std::vector<Feature> features; // Current features being tracked
    int feature_id_counter = 0;    // Counter to assign unique IDs

    // Parameters for feature detection
    const int maxFeatures = 200; // Maximum number of features to track
    const float32_t qualityLevel = 100.F;
    const float32_t minDistance = 15.F;
    const int blockSize = 3;
    const bool useHarrisDetector = false;
    const double k = 0.04;

    // Running sum variable
    uint64_t running_sum = 0; // Use a larger type to prevent overflow

    // Iterate through each PNG file
    for (size_t i = 0; i < pngFiles.size(); ++i) {
        // Read the current image in grayscale
        cv::Mat img;
        img = cv::imread(pngFiles[i].string(), cv::IMREAD_GRAYSCALE);
        cv::equalizeHist(img, img);

        if (img.empty()) {
            std::cerr << "Failed to read image: " << pngFiles[i] << std::endl;
            continue;
        }

        // Assign to current grayscale image
        currGray = img;
        int width = img.cols;
        int height = img.rows;
        size_t stride = img.step;
        if (stride % 8 != 0) {
            printf("stride should be multiple of 8\n");
        }

        uint64_t start = mcv_time_monotonic_ns();
        
        
        // If not the first frame, perform tracking
        if (!prevGray.empty() && !features.empty()) {
            // Prepare points for tracking
            std::vector<Point> prevPoints;
            std::vector<int> featureIDs;
            for (const auto& feat : features) {
                prevPoints.push_back(feat.point);
                featureIDs.push_back(feat.id);
            }

            bool free_src1_data = false;
            unsigned char* src1_data = getAlignedData(prevGray, free_src1_data);
            if (src1_data == nullptr) {
                std::cerr << "Failed to allocate aligned memory for src1_data.\n";
                return -1;
            }

            bool free_src2_data = false;
            unsigned char* src2_data = getAlignedData(currGray, free_src2_data);
            if (src2_data == nullptr) {
                std::cerr << "Failed to allocate aligned memory for src2_data.\n";
                return -1;
            }

            int n_pyr_levels = 3;
            fcvPyramidLevel* src1_pyr = (fcvPyramidLevel*)malloc(3 * sizeof(fcvPyramidLevel));
            fcvPyramidLevel* src2_pyr = (fcvPyramidLevel*)malloc(3 * sizeof(fcvPyramidLevel));

            fcvPyramidAllocate(src1_pyr, width, height, sizeof(uint8_t), n_pyr_levels, 0);
            fcvPyramidAllocate(src2_pyr, width, height, sizeof(uint8_t), n_pyr_levels, 0);
            
            fcvPyramidCreateu8(src1_data, width, height, n_pyr_levels, src1_pyr);
            fcvPyramidCreateu8(src2_data, width, height, n_pyr_levels, src2_pyr);

            /*
            for(int j = 0; j < n_pyr_levels; j++) {
                printf("lvl: %d, w: %d, h: %d\n", j, src1_pyr[j].width, src1_pyr[j].height);
                cv::Mat lvl_img(src1_pyr[j].height, src1_pyr[j].width, CV_8UC1, (void*)src1_pyr[j].ptr);
                std::string pyr_out = "/data/benchmark-results/pyr/" + std::to_string(j) + ".png";
                cv::imwrite(pyr_out, lvl_img);
            }
            */

            size_t xy_size = 2 * features.size() * sizeof(float);
            float* xy = (float*)malloc(xy_size);
            float* xy_out = (float*)malloc(xy_size);

            size_t status_size = features.size() * sizeof(uint32_t);
            int32_t* status = (int32_t*)malloc(status_size);

            for(int j = 0; j < features.size(); j++) {
                xy[j * 2]       = (float)features[j].point.x;
                xy[j * 2 + 1]   = (float)features[j].point.y;
            }


            fcvTrackLKOpticalFlowf32(   src1_data,
                                        src2_data,
                                        width,
                                        height,
                                        src1_pyr,
                                        src2_pyr,
                                        NULL,
                                        NULL,
                                        xy,
                                        xy_out,
                                        status,
                                        features.size(),
                                        9,
                                        9,
                                        7,
                                        n_pyr_levels,
                                        0, 0, 0, 0 );

            fcvPyramidDelete(src1_pyr, n_pyr_levels, 1);
            fcvPyramidDelete(src2_pyr, n_pyr_levels, 1);

            free(src1_pyr);
            free(src2_pyr);


            // Temporary storage for updated features
            std::vector<Feature> prev_features;
            std::vector<Feature> updatedFeatures;

            // patch comes from the prev img
            int patch_size = 8;
            uint8_t* patch = (uint8_t*)malloc(patch_size * patch_size * sizeof(uint8_t));

            for (size_t j = 0; j < features.size(); ++j) {
                if (status[j] == 1) { // Successfully tracked
                    Feature feat;
                    feat.id = featureIDs[j];
                    float new_x = xy_out[j * 2];
                    float new_y = xy_out[j * 2 + 1];
                    feat.point = (Point){new_x, new_y};
                    feat.age = features[j].age + 1;

                    uint16_t center_x = static_cast<uint16_t>(std::round(features[j].point.x));
                    uint16_t center_y = static_cast<uint16_t>(std::round(features[j].point.y));

                    uint16_t best_x = 0;
                    uint16_t best_y = 0;
                    uint32_t bestNCC = 0;
                    float subX = 0.0f;
                    float subY = 0.0f;
                    
                    /*
                    if (center_x < 4 || center_x >= width-3 || center_y < 4 || center_y >= height - 3) {
                        printf("pixel is too close to the border.\n");
                    } else {
                        
                        for (int dy = -4; dy <= 3; ++dy) {
                            for (int dx = -4; dx <= 3; ++dx) {
                                int src_x = center_x + dx;
                                int src_y = center_y + dy;

                                int patch_x = dx + 4;
                                int patch_y = dy + 4;

                                patch[patch_y * patch_size + patch_x] = src1_data[src_y * patch_size + src_x];
                            }
                        }

                        fcvNCCPatchOnCircle8x8u8(   patch,
                                                    src2_data,
                                                    width,
                                                    height,
                                                    static_cast<unsigned short>(std::round(new_x)),
                                                    static_cast<unsigned short>(std::round(new_y)),
                                                    5,
                                                    &best_x,
                                                    &best_y,
                                                    &bestNCC,
                                                    1,
                                                    &subX,
                                                    &subY );

                        // printf("pred pos: (%.2f, %.2f), best coord: (%4d, %4d) sub (%.2f, %.2f), ncc: %d\n", 
                        //         new_x, new_y, best_x, best_y, 
                        //         (float)best_x + subX, (float)best_y + subY, static_cast<int32_t>(bestNCC));

                        if (bestNCC < 80) {
                            // feat.point = (Point){   (float)best_x + subX, 
                            //                         (float)best_y + subY };
                            continue;
                        }
                    }
                    */

                    updatedFeatures.push_back(feat);
                    prev_features.push_back(features[j]);
                } else {
                    printf("not tracked: %d\n", status[j]);
                }
            }

            free(patch);


            free(xy);
            free(xy_out);
            free(status);


            // the src is the current img, and the search (x,y) is the predicted
            // loc that comes from the optic flow calc
            




            if (free_src1_data) free(src1_data);
            if (free_src2_data) free(src2_data);
            

            // std::vector<bool> has_low_gradient = isLowGradient(currGray, updatedFeatures); 

            // Replace old features with updated ones
            // features.clear();
            // for (int j = 0; j < updatedFeatures.size(); j++) {
            //     if (!has_low_gradient[j]) {
            //         features.push_back(updatedFeatures[j]);
            //     }
            // }

            /*
            double nccThreshold = 0.3;
            int patchSize = 5;

            // NCC checking
            std::vector<Feature> trackingResults;
            int halfPatch = patchSize / 2;

            for (size_t j = 0; j < prev_features.size(); ++j) {
                cv::Point2f prevPt = cv::Point2f(prev_features[j].point.x, prev_features[j].point.y);
                cv::Point2f nextPt = cv::Point2f(updatedFeatures[j].point.x, updatedFeatures[j].point.y);

                // Define patch regions, ensuring they are within image bounds
                if (prevPt.x - halfPatch < 0 || prevPt.x + halfPatch >= prevGray.cols ||
                    prevPt.y - halfPatch < 0 || prevPt.y + halfPatch >= prevGray.rows ||
                    nextPt.x - halfPatch < 0 || nextPt.x + halfPatch >= currGray.cols ||
                    nextPt.y - halfPatch < 0 || nextPt.y + halfPatch >= currGray.rows) {
                    continue;
                }

                // Extract patches around the points
                cv::Rect prevRect(prevPt.x - halfPatch, prevPt.y - halfPatch, patchSize, patchSize);
                cv::Rect nextRect(nextPt.x - halfPatch, nextPt.y - halfPatch, patchSize, patchSize);
                cv::Mat prevPatch = prevGray(prevRect);
                cv::Mat nextPatch = currGray(nextRect);

                // Calculate NCC
                double nccValue = calculateNCC(prevPatch, nextPatch);

                // Determine if the feature was successfully tracked based on NCC threshold
                if (nccValue >= nccThreshold) {
                    // printf("ncc: %f\n", nccValue);
                    trackingResults.push_back(updatedFeatures[j]); // Successfully tracked
                } else {
                    // Poorly tracked
                }
            }
            */

            features = updatedFeatures;
        }

        // Detect new features if needed
        if (features.size() < maxFeatures) {
            // Convert to a mask to avoid detecting existing features
            // cv::Mat mask = cv::Mat::ones(currGray.size(), CV_8UC1) * 255;
            // for (const auto& feat : features) {
            //     cv::circle(mask, feat.point, (int)minDistance, cv::Scalar(0), -1);
            // }

            // Detect new features
            std::vector<Point> newPoints;


            bool free_src_data = false;
            unsigned char* src_data = getAlignedData(currGray, free_src_data);
            if (src_data == nullptr) {
                std::cerr << "Failed to allocate aligned memory for srcData.\n";
                return -1;
            }

            uint32_t* xy = nullptr;
            size_t xySize = 2 * maxFeatures * sizeof(uint32_t);
            if (posix_memalign(reinterpret_cast<void**>(&xy), 16, xySize) != 0) {
                std::cerr << "Failed to allocate aligned memory for xy.\n";
                return -1;
            }

            uint32_t num_features = 0;

            fcvGoodFeatureToTracku8(    src_data,
                                        width,
                                        height,
                                        stride,
                                        minDistance,
                                        20,
                                        qualityLevel,
                                        xy,
                                        maxFeatures - features.size(),
                                        &num_features );


            // Add new features with unique IDs and age 1
            // printf("points found: %d\n", num_features);
            for (int j = 0; j < num_features; j++) {
                float x = xy[j * 2];
                float y = xy[j * 2 + 1];
                cv::Point2f newPoint(x, y);

                bool isFarEnough = true;
                for (const auto& f : features) {
                    if (calculateDistance(newPoint, cv::Point2f(f.point.x, f.point.y)) < minDistance) {
                        isFarEnough = false;
                        break;
                    }
                }

                if (isFarEnough) {
                    Feature feat;
                    feat.id = feature_id_counter++;
                    feat.point = (Point){x, y};
                    feat.age = 1;
                    features.push_back(feat);
                }
            }
            
            free(xy);
            if (free_src_data) {
                free(src_data);
            }
        }

        std::vector<Feature> temp_features;

        // calculateShiTomasiScores(currGray, features);
        // std::sort(features.begin(), features.end(), [](const Feature& a, const Feature& b) {
        //     if (a.age != b.age) {
        //         return a.age > b.age; // Sort by age first (descending)
        //     } 
        //     else {
        //         return a.id > b.id; // If ages are equal, sort by score (descending)
        //     }
        // });

        for (const auto& feat : features) {
            if (feat.score > 0) {
                temp_features.push_back(feat);
            }
        }

        // features = temp_features;



        // After updating features, find the top 30 oldest
        std::vector<Feature> topOldestFeatures = getTopNOldestFeatures(features, 30);

        // Calculate and update running_sum with the ages of top 30 features
        for (const auto& feat : topOldestFeatures) {
            running_sum += feat.age;
        }

        uint64_t end = mcv_time_monotonic_ns();
        printf("Frame: %4d, Iter Time ::    %6.2fms\n", i, (double)(end - start) / 1000000);



        // Output the top 30 oldest features for the current frame
        /*
        std::cout << "Frame " << i + 1 << ": Top 30 Oldest Features" << std::endl;
        std::cout << "ID\tAge\tPosition (x, y)" << std::endl;
        for (const auto& feat : topOldestFeatures) {
            std::cout << feat.id << "\t" << feat.age << "\t(" 
                      << std::fixed << std::setprecision(2) << feat.point.x 
                      << ", " << feat.point.y << ")" << std::endl;
        }
        std::cout << "-----------------------------------------" << std::endl;
        */

        // Optionally, visualize the tracking
        cv::Mat vis;
        cv::cvtColor(currGray, vis, cv::COLOR_GRAY2BGR);
        for (const auto& feat : features) {
            cv::Point2f cv_point(feat.point.x, feat.point.y);
            cv::circle(vis, cv_point, 3, cv::Scalar(0, 255, 0), -1);
            cv::putText(vis, std::to_string(feat.id), cv_point, 
                        cv::FONT_HERSHEY_SIMPLEX, 0.4, cv::Scalar(0, 0, 255), 1);
        }
        std::string out_path = "/data/benchmark-results/test/" + std::to_string(i) + ".png";
        cv::imwrite(out_path, vis);
        // cv::waitKey(1);

        // printf("features size: %d\n", features.size());
        

        // Update previous image
        prevGray = currGray.clone();
    }

    // After processing all frames, print the running_sum
    std::cout << "Total Running Sum of Ages of Top 30 Oldest Features: " << running_sum << std::endl;

    return 0;
}
