/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

/**
 * @file        mcv-test-ncc.cpp
 *
 * @brief       Unit test for NCC module of modalcv
 *
 * @author      Thomas Patton (thomas.patton@modalai.com) 
 * @date        2023 
 */

#include <fcntl.h>
#include <getopt.h>
#include <modalcv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>   // for mkdir
#include <sys/types.h>  // for mode_t in mkdir
#include <time.h>
#include <unistd.h>

#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>

using namespace cv;

// input files
// #define TEST_SRC_IMAGE "/usr/share/libmodalcv/dfs/left.png"
// #define TEST_REF_IMAGE "/usr/share/libmodalcv/dfs/right.png"

#define TEST_SRC_IMAGE "/usr/share/libmodalcv/ov9782/left.jpg"
#define TEST_REF_IMAGE "/usr/share/libmodalcv/ov9782/right.jpg"

// output files
#define OUTPUT_FOLDER "/data/libmodalcv/ncc/"
#define LOOPS 100  // run test 100 times for timing purposes

static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
        fprintf(stderr, "ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec * 1000000000 + (int64_t)ts.tv_nsec;
}

int main() {
    mcv_en_debug_prints(1);

    // make output directory to save debug images to
    if (mkdir(OUTPUT_FOLDER, S_IRWXU) && errno != EEXIST) {
        fprintf(stderr, "failed to make folder\n");
        return -1;
    }

    // load in images
    printf("loading images from disk\n");
    Mat src_img, ref_img;
    src_img = imread(TEST_SRC_IMAGE, IMREAD_GRAYSCALE);
    ref_img = imread(TEST_REF_IMAGE, IMREAD_GRAYSCALE);

    int w = src_img.cols;
    int h = src_img.rows;

    if (ref_img.type() != CV_8UC1 || src_img.type() != CV_8UC1) {
        fprintf(stderr, "ERROR loading images\n");
        return -1;
    }

    // configure and run feature extraction to have points for NCC
    mcv_cvp_fpx_config_t fpx_config;
    fpx_config.width           = w;
    fpx_config.height          = h;
    fpx_config.mode            = MCV_CVP_FPX_PEAK_8x8;
    fpx_config.nms_mode        = MCV_CVP_FPX_5_TAP_NMS;
    fpx_config.score_threshold = 250;
    fpx_config.robustness      = 50;                     // 0-127, default 10

    // get a handle for our FPX process
    mcv_cvp_fpx_handle fpx_handle = mcv_cvp_fpx_init(fpx_config);
    if(fpx_handle == NULL) {
        fprintf(stderr, "ERROR getting FPX handle\n");
        return -1;
    }

    // create output object
    mcv_cvp_fpx_t fpx;
    if(mcv_cvp_create_fpx(fpx_handle, &fpx)) {
        fprintf(stderr, "ERROR creating FPX output object\n");
    }

    // extract features
    printf("extracting features\n");
    for(int i = 0; i < LOOPS; i++) {
        if(mcv_cvp_fpx_process(fpx_handle, src_img.data, &fpx))
        {
            fprintf(stderr, "ERROR completing FPX process\n");
            return -1;
        }
    }
    printf("n_points:  %5d\n", fpx.n_points);
    printf("max_score: %5d\n", fpx.max_score);

    // configure NCC
    mcv_cvp_ncc_config_t config;
    config.width         = w;
    config.height        = h;
    config.ncc_mode      = MCV_CVP_NCC_FRAME_MODE;
    config.score_type    = MCV_CVP_NCC_ALL_SCORE;
    config.en_robustness = 0;
    config.en_subpel     = 1;

    // initialize NCC and get handle
    mcv_cvp_ncc_handle handle = mcv_cvp_ncc_init(config);
    if(NULL == handle) {
        fprintf(stderr, "ERROR obtaining handle\n");
        return -1;
    }

    // create NCC output structure
    mcv_cvp_ncc_output_t output;
    mcv_cvp_create_ncc_output(handle, &output);

    // now run NCC process
    mcv_cvp_ncc_process(handle, src_img.data, ref_img.data, &fpx, &output);

    printf("NCC Process complete\n");

    // create a canvas
    cv::Mat final_img;
    cv::hconcat(src_img, ref_img, final_img);
    cv::cvtColor(final_img, final_img, cv::COLOR_GRAY2BGR);

    // draw NCC matches
    int match_radius = 4;
    for (int i = 0; i < output.n_results; i++) {
        // draw original features
        // cv::drawMarker(final_img, cv::Point2f(fpx.features[i].x, fpx.features[i].y),
        //     cv::Scalar(0,255,0), cv::MARKER_SQUARE, 4, 1);
        
        // if there's a match, draw it!
        if(output.valid_match[i]) {
            printf("i: %d, score: %d\n", i, output.scores[i]);
            int shifted_x = output.results[i].match_loc_x + w; 
            int shifted_y = output.results[i].match_loc_y; 

            cv::rectangle(final_img, cv::Point2f(fpx.features[i].x-4,
                fpx.features[i].y-4), cv::Point2f(fpx.features[i].x+4,
                fpx.features[i].y+4), cv::Scalar(0,255,0), 1);
            cv::rectangle(final_img, cv::Point2f(w + fpx.features[i].x-9,
                fpx.features[i].y-9), cv::Point2f(w + fpx.features[i].x+9,
                fpx.features[i].y+9), cv::Scalar(0,255,255), 1);

            // draw a box around the match location
            cv::rectangle(final_img, cv::Point2f(shifted_x-match_radius, 
                shifted_y-match_radius), cv::Point2f(shifted_x+match_radius,
                shifted_y+match_radius), cv::Scalar(0,0,255), 1);
        }
    }
    printf("done drawing, now writing images\n");
    cv::imwrite("/data/libmodalcv/ncc/ncc_out.png", final_img);

    // cleanup
    free(fpx.features);
    mcv_cvp_fpx_deinit(fpx_handle);
    mcv_cvp_destroy_output(&output);
    mcv_cvp_ncc_deinit(handle);
    return 0;
}
