/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file       mcv-test-opt-flow.cpp 
 * 
 *             This test is intended to be run with a path to some voxl-logger
 *             dataset as an argument. 
 *
 * @brief      Test of optic flow on CVP hardware
 * 
 * @author     Thomas Patton (thomas.patton@modalai.com)
 * @date       2024
 */

#include "modalcv/opt_flow.h"
#include <fcntl.h>
#include <getopt.h>
#include <modalcv.h>
#include <modalcv/io.hpp>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>  // for mkdir
#include <sys/types.h> // for mode_t in mkdir
#include <time.h>
#include <unistd.h>
#include <iostream>
#include <iomanip>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include <sstream>

#define OUTPUT_FOLDER "/data/libmodalcv/opt_flow/"

/**
 * @brief Gets the system time
 * 
 * @return int64_t The system time
 */
static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
        fprintf(stderr, "ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec * 1000000000 + (int64_t)ts.tv_nsec;
}

/**
 * @brief A method for drawing the output of our optic flow method
 */
static void draw_opt_flow(
    cv::Mat                                           img,
    mcv_cvp_motion_vector_t*                          motion_vector,
    __attribute__((unused)) mcv_cvp_opt_flow_stats_t* stats,
    int                                               w,
    int                                               h) {
    int n_rows = h / 8;
    int n_cols = w / 8; // 8x8 block

    for (int row_i = 0; row_i < n_rows; row_i++) {
        for (int col_i = 0; col_i < n_cols; col_i++) {
            int arrow_start_x = (col_i * 8 + 4);
            int arrow_start_y = (row_i * 8 + 4);
            int coord         = x_y_to_flat(col_i, row_i, n_cols, n_rows, 1);
            int arrow_end_x   = (arrow_start_x - motion_vector[coord].mv_x);
            int arrow_end_y   = (arrow_start_y - motion_vector[coord].mv_y);

            // max variance -> min part
            cv::line(
                img,
                cv::Point2f(arrow_start_x, arrow_start_y),
                cv::Point2f(arrow_end_x, arrow_end_y),
                cv::Scalar(0, 255, 255),
                1,
                1);
        }
    }
}

int main(int argc, char* argv[]) {
    mcv_en_debug_prints(1);
    mcv_en_timing_prints(1);

    // vars to hold args
    int w, h;
    std::vector<std::string> paths;

    // parse input arguments and extract paths for optic flow
    if(init_modalcv_example(argc, argv, paths, w, h)) {
        exit(1);
    }

    // clear output directory
    if (std::system("mkdir -p /data/libmodalcv/opt_flow")) {
        std::cout << "Error creating directories" << std::endl;
        return -1;
    }
    if (std::system("rm -rf /data/libmodalcv/opt_flow/*")) {
        fprintf(
            stderr,
            "Warning, failed to delete old files from %s\n",
            OUTPUT_FOLDER);
        return -1;
    }

    // configure optic flow process
    mcv_cvp_opt_flow_config_t config;
    config.width      = w;
    config.height     = h;
    config.input_ubwc = 0;
    config.en_stats   = 0;

    // get a handle for optical flow
    mcv_cvp_opt_flow_handle handle = mcv_cvp_opt_flow_init(config);
    if (NULL == handle) {
        fprintf(stderr, "ERROR getting opt flow handle\n");
        return -1;
    }

    // create an output object
    mcv_cvp_opt_flow_output_t output;
    mcv_cvp_create_opt_flow_output(handle, &output);

    // iterate over images as requested and run opt flow
    printf("starting main loop...\n");
    int initialized = 0;
    for (size_t i = 1; i < paths.size(); i++) { // note opt flow needs to start from 1
        cv::Mat curr_img_gray, curr_img_color;
        std::cout << paths[i] << std::endl;
        curr_img_gray = cv::imread(paths[i], cv::IMREAD_GRAYSCALE);
        cv::cvtColor(curr_img_gray, curr_img_color, cv::COLOR_GRAY2BGR);
        int64_t start = _apps_time_monotonic_ns();
        if (initialized) {
            // run opt flow using current img and img data in memory
            mcv_cvp_opt_flow_process_from_mem(
                handle, curr_img_gray.data, &output);

        } else {
            cv::Mat prev_img_gray, prev_img_color;
            prev_img_gray = cv::imread(paths[i - 1], cv::IMREAD_GRAYSCALE);
            cv::cvtColor(prev_img_gray, prev_img_color, cv::COLOR_GRAY2BGR);

            // run opt flow using prev and current img
            mcv_cvp_opt_flow_process_from_img(
                handle, prev_img_gray.data, curr_img_gray.data, &output);

            // have image into mem now
            initialized = 1;
        }
        int64_t end = _apps_time_monotonic_ns();
        printf("Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

        // draw optic flow vectors and write to disk
        std::stringstream ss;
        ss << std::setw(5) << std::setfill('0') << i;
        std::string output_path = OUTPUT_FOLDER + ss.str() + ".png";
        draw_opt_flow(
            curr_img_color,
            output.mvs,
            output.stats,
            curr_img_gray.cols,
            curr_img_gray.rows);
        cv::imwrite(output_path, curr_img_color);
    }

    // cleanup
    printf("all done, deinitializing\n");
    mcv_cvp_destroy_opt_flow_output(handle, &output);
    mcv_cvp_opt_flow_deinit(handle);
    return 0;
}
