/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file       mcv-test-pyr-fpx.cpp
 *
 * @brief      Tests the PYR Feature Extraction (FPX) process
 *
 * @author     Kyle Tyni (kyle.tyni@modalai.com)
 * @date       2024
 */


#include <fastcv.h>
#include <fcntl.h>
#include <getopt.h>
#include <modalcv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>   // for mkdir
#include <sys/types.h>  // for mode_t in mkdir
#include <time.h>
#include <unistd.h>
#include <cmath>

#include <experimental/filesystem>
#include <iomanip>
#include <iostream>
#include <modal_pipe.h>
#include <unordered_map>
#include <sstream>
#include <voxl_common_config.h>
#include <random>
#include <modalcv/io.hpp>
#include "modalcv/pyr.h"

#include "../src/cvp/internal/private_types.h"


#include <cvpMem.h>
#include <cvpPyramid.h>
#include <cvpSession.h>
#include <cvpTypes.h>

namespace fs = std::experimental::filesystem;

#define PROCESS_NAME "mcv-test-pyr-fpx"

// vio cam config from common vio_cams.conf
#define VFT_MAX_GROUP_CAMS      4

int n_cams;
vio_cam_t vio_cams[VFT_MAX_GROUP_CAMS];


/**
 * @brief Gets the system time
 * 
 * @return int64_t The system time
 */
static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
        fprintf(stderr, "ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec * 1000000000 + (int64_t)ts.tv_nsec;
}



/**
 * @brief Exits the program cleanly
 */
static void _quit(int ret)
{
	// Close all the open pipe connections
	pipe_server_close_all();
	pipe_client_close_all();


	// remove this process ID file
	remove_pid_file(PROCESS_NAME);

	if (ret == 0)
		printf("Exiting Cleanly\n");
	else
		printf("error code %d\n", ret);
	exit(ret);
	return;
}

mcv_cvp_pyr_fpx_handle pyr_fpx_handle[3];
mcv_cvp_pyr_fpx_t output_pyr_fpx[3];

int initialized[3];

void _camera_helper_cb(__attribute__((unused)) int ch, camera_image_metadata_t meta, 
                                            char *frame, void *context) {

    pyr_fpx_session_t* session1 = (pyr_fpx_session_t*)pyr_fpx_handle[ch];

    memcpy(session1->input.pBuffer->pAddress, frame, meta.width * meta.height);

    if (CVP_SUCCESS != cvpPyramid_Sync( session1->handle, 
                                        &session1->input,
                                        session1->config.score_shift,
                                        session1->config.score_threshold,
                                        session1->config.robustness,
                                        (cvpFpxNmsTap)session1->config.nms_mode,
                                        &session1->output)) {
        fprintf(stderr, "ERROR computing Image Pyaramids + FPX\n");
        return; 
    }
    
    return;
}


static void _camera_disconnect_cb(__attribute__((unused)) int ch,
								  __attribute__((unused)) void *context) {
	// TODO remove from active cams list so we can keep publishing is one cam crashes
	fprintf(stderr, "Disconnected from camera pipe\n");
}



int w = 1280;
int h = 800;


int main(int argc, char* argv[]) {
	mcv_en_debug_prints(0);
	mcv_en_timing_prints(0);


	// load only enabled vio cams from common vio-cam config file
	// no need to check for extrinsics or cam cal, we don't need those
	n_cams = vcc_read_vio_cam_conf_file(vio_cams, VFT_MAX_GROUP_CAMS, 1);
	if(n_cams<1) return -1;
	printf("vio_cam config:\n");
	vcc_print_vio_cam_conf(vio_cams, n_cams);
	printf("=================================================================\n");


	/* make sure another instance isn't running
	 * if return value is -3 then a background process is running with
	 * higher privileges and we couldn't kill it, in which case we should
	 * not continue or there may be hardware conflicts. If it returned -4
	 * then there was an invalid argument that needs to be fixed.
	 */
	if(kill_existing_process(PROCESS_NAME, 2.0) < -2) {
		fprintf(stderr, "ERROR: could not kill existing process\n");
		_quit(-1);
	}

	// start signal handler so we can exit cleanly
	if(enable_signal_handler() == -1) {
		fprintf(stderr, "ERROR: failed to start signal handler\n");
		_quit(-1);
	}

	// set up affinity for this process
	set_cpu_affinity(cpu_set_big_cores());
	// set_cpu_affinity(cpu_set_big_cores_and_gold_core());
	// print_cpu_affinity();
	pipe_set_process_priority(THREAD_PRIORITY_RT_HIGH);

	// Set the main running flag to 1 to indicate that we are running
	main_running = 1;


	/* make PID file to indicate your project is running
	 * due to the check made on the call to rc_kill_existing_process() above
	 * we can be fairly confident there is no PID file already and we can
	 * make our own safely.
	 */
	make_pid_file(PROCESS_NAME);

	cvp_set_pixel_search_radius(10);
	cvp_set_octave_search_radius(3);
	cvp_set_en_id_sort(1);



	// for each input, open the camera pipe
	for(int i=0; i<n_cams; i++) {

		pipe_client_set_camera_helper_cb(i, _camera_helper_cb, NULL);
		pipe_client_set_disconnect_cb(i, _camera_disconnect_cb,NULL);
		
        int flags = CLIENT_FLAG_EN_CAMERA_HELPER;
		printf("opening pipe: %s\n", vio_cams[i].pipe_for_tracking);

		if(pipe_client_open(i, vio_cams[i].pipe_for_tracking, PROCESS_NAME, flags, 0)) {
			printf("FAILED TO OPEN %s\n", vio_cams[i].pipe_for_tracking);
			return -1;
		}
	}


    for (int i = 0; i < n_cams; i++) {

        // init pyr fpx flow config
        mcv_cvp_pyr_fpx_config_t pyr_fpx_config = mcv_cvp_pyr_fpx_default_config();
        pyr_fpx_config.width = w;
        pyr_fpx_config.height = h;

        // create a pyr_fpx handle from our configuration
        pyr_fpx_handle[i] = mcv_cvp_pyr_fpx_init(pyr_fpx_config);
        if (NULL == pyr_fpx_handle[i]) {
            fprintf(stderr, "ERROR creating pyr fpx handle\n");
        }

        // create an output object
        if (mcv_cvp_create_pyr_fpx(pyr_fpx_handle[i], &output_pyr_fpx[i])) {
            fprintf(stderr, "ERROR creating feature extraction pyramid\n");
            return -1;
        }
    }


	// run until start/stop module catches a signal and changes main_running to 0
	printf("Setup completed! Waiting for images...\n");
	while (main_running) usleep(500000);

	// shutdown nicely
	_quit(0);

    for (int i = 0; i < n_cams; i++) {

        // pyramid feature extraction cleanup
        mcv_cvp_destroy_pyr_fpx(pyr_fpx_handle[i], &output_pyr_fpx[i]);
        mcv_cvp_pyr_fpx_deinit(pyr_fpx_handle[i]);
    }


    return 0;
}