/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/stat.h>	// for mkdir
#include <sys/types.h>	// for mode_t in mkdir
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <iomanip>
#include <iostream>
#include <modalcv.h>
#include <modalcv/io.hpp>

#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>


#define OUTPUT_FOLDER	"/data/libmodalcv/pyr/"

const std::string OCTAVE_FOLDERS[] = {
    std::string(OUTPUT_FOLDER) + "octave_0/",
    std::string(OUTPUT_FOLDER) + "octave_1/",
    std::string(OUTPUT_FOLDER) + "octave_2/"
};


/**
 * @brief Helper fn for getting the current clock time in nanoseconds
 *
 * @return Time in ns
 */
static int64_t _apps_time_monotonic_ns() {
	struct timespec ts;
	if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
		fprintf(stderr,"ERROR calling clock_gettime\n");
		return -1;
	}
	return (int64_t)ts.tv_sec*1000000000 + (int64_t)ts.tv_nsec;
}


int main(int argc, char* argv[]) {
	mcv_en_debug_prints(0);
	mcv_en_timing_prints(1);
	
	int w, h;
    std::vector<std::string> paths;

    // args checks, makes sure path is good
    if (init_modalcv_example(argc, argv, paths, w, h)) {
        fprintf(stderr, "ERROR, failed to initialize example\n");
        return -1;
    }

	if (std::system("mkdir -p /data/libmodalcv/pyr")) {
        std::cout << "Error creating directories" << std::endl;
        return -1;
    }
    if (std::system("rm -rf /data/libmodalcv/pyr/*")) {
        fprintf(
            stderr,
            "Warning, failed to delete old files from %s\n",
            OUTPUT_FOLDER);
        return -1;
    }

	for (int i = 0; i < 3; ++i) {
        // Create directories
		std::string mkdirCommand = "mkdir -p " + OCTAVE_FOLDERS[i];
        if (std::system(mkdirCommand.c_str())) {
            std::cout << "Error creating directory: " << OCTAVE_FOLDERS[i] << std::endl;
            return -1;
        }
        
        // Remove old files
		std::string rmCommand = "rm -rf " + OCTAVE_FOLDERS[i] + "*";
        if (std::system(rmCommand.c_str())) {
            std::cerr << "Warning, failed to delete old files from " << OCTAVE_FOLDERS[i] << std::endl;
            return -1;
        }
    }

	// initialize pyramid extractor
	mcv_cvp_pyr_config_t pyr_config;
	pyr_config.width               = w;
	pyr_config.height              = h;
	pyr_config.n_octaves           = 3;
	pyr_config.n_scales_per_octave = 1;

	// get a handle for our PYR downscaler process
	mcv_cvp_pyr_handle handle = mcv_cvp_pyr_init(pyr_config);
	if(NULL == handle) {
		fprintf(stderr, "ERROR creating pyramid handle\n");
		return -1;
	}

	// create output structure
	mcv_cvp_pyr_t pyr;
	if (mcv_cvp_create_pyr(handle, &pyr)) {
		fprintf(stderr, "ERROR creating PYR output object\n");
		return -1;
	}

	// iterate over all images and perform pyramid extraction
	printf("extracting image pyramid\n");
	for (size_t i = 0; i < paths.size(); i++) {

        cv::Mat img_mat = cv::imread(paths[i], cv::IMREAD_GRAYSCALE);

		if (img_mat.type()!= CV_8UC1) {
            fprintf(stderr, "ERROR loading %s\n", paths[i].c_str());
            return -1;
        }

		int64_t start = _apps_time_monotonic_ns();
		// do the pyramid extraction on an individual image
		if(mcv_cvp_pyr_process(handle, img_mat.data, &pyr)) {
			fprintf(stderr, "ERROR executing pyramid process\n");
			return -1;
		}

		int64_t end = _apps_time_monotonic_ns();
		printf("Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

		// save to disk
		for (int j = 0; j < (int)pyr_config.n_octaves; j++) {
        	std::stringstream ss;
			ss << std::setw(5) << std::setfill('0') << i;
			std::string output_path = OCTAVE_FOLDERS[j] + ss.str() + ".png";
			printf("%s\n", output_path.c_str());

			cv::Mat curr_pyr(pyr.height[j], pyr.width[j], CV_8UC1, 
							 pyr.image_pyramids[j]);

			cv::imwrite(output_path, curr_pyr);
		}
	}

	// ////////////////////////////////////////////////////////////////////////////
	// cleanup
	////////////////////////////////////////////////////////////////////////////
	if(mcv_cvp_destroy_pyr(handle, &pyr))
	{
		fprintf(stderr, "ERROR destorying pyramid object\n");
		return -1;
	}
	if(mcv_cvp_pyr_deinit(handle))
	{
		fprintf(stderr, "ERROR deinitializing pyramid\n");
		return -1;
	}

	printf("DONE\n");
	return 0;
}
