/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

/**
 * @file       mcv-test-undistort.cpp <examples/mcv-test-undistort>
 *
 * @brief      A few unit tests of the undistortion process on CPU and CVP
 * 			   hardware.
 * 
 * @author     Thomas Patton (thomas.patton@modalai.com)
 * @date       2023
 */

#include <stdio.h>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/stat.h>	// for mkdir
#include <sys/types.h>	// for mode_t in mkdir
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <modalcv.h>

#include <opencv2/imgcodecs.hpp>
using namespace cv;

#define TEST_IMAGE_1			"/usr/share/libmodalcv/tracking/00000.png"
#define INTRINSICS_FILE_1		"/usr/share/libmodalcv/tracking/opencv_tracking_intrinsics.yml"
#define OUTPUT_FILE_1_CPU		"/data/libmodalcv/tracking_rectified_cpu.png"
#define OUTPUT_FILE_1_CVP		"/data/libmodalcv/tracking_rectified_cvp.png"

#define TEST_IMAGE_2			"/usr/share/libmodalcv/J2119/00900.png"
#define INTRINSICS_FILE_2		"/usr/share/libmodalcv/J2119/modalai/opencv_trackingL_intrinsics.yml"
#define OUTPUT_FILE_2_CPU		"/data/libmodalcv/J2119_cpu.png"
#define OUTPUT_FILE_2_CVP		"/data/libmodalcv/J2119_cvp.png"

#define OUTPUT_FOLDER	        "/data/libmodalcv/"
#define OUTPUT_DIFF		        "/data/libmodalcv/diff.png"

#define LOOPS	100 // run test 100 times for timing purposes

static int64_t _apps_time_monotonic_ns()
{
	struct timespec ts;
	if(clock_gettime(CLOCK_MONOTONIC, &ts)){
		fprintf(stderr,"ERROR calling clock_gettime\n");
		return -1;
	}
	return (int64_t)ts.tv_sec*1000000000 + (int64_t)ts.tv_nsec;
}

/**
 * @brief Generates the difference of two images and saves it
 * 
 * @param image_path_1 Path to the first image to take the difference of
 * @param image_path_2 Path to the second image to take the difference of
 * @param output_path Where to save the difference image
 * @return 0 if successful, -1 otherwise
 */
static int generate_diff(const char* image_path_1, const char* image_path_2, const char* output_path)
{
	// load both images
	Mat img_mat_1;
	img_mat_1 = imread(image_path_1, IMREAD_GRAYSCALE);
	if(img_mat_1.type()!=CV_8UC1){
		fprintf(stderr, "ERROR loading %s\n", image_path_1);
		return -1;
	}
	Mat img_mat_2;
	img_mat_2 = imread(image_path_2, IMREAD_GRAYSCALE);
	if(img_mat_2.type()!=CV_8UC1){
		fprintf(stderr, "ERROR loading %s\n", image_path_2);
		return -1;
	}

	// signed int, times 10, plus 128, bounded by opencv on imwrite
	Mat diff = 10*(img_mat_1 - img_mat_2) + 128;

	imwrite(output_path, diff);
	return 0;
}

/**
 * @brief Does LOOPS number of iterations to determine the speed of the warping process on CPU
 * 
 * @param img_mat The CV matrix to warp
 * @param intrinsics_file String location of the intrinsics file for the camera
 * @param output A buffer for the undistorted image
 * @return 0 if successful, -1 otherwise
 */
static int run_test_cpu(Mat img_mat, const char* intrinsics_file, uint8_t* output)
{
	int ret;

	// configure lookup table for warp
	mcv_undistort_map_t map = MCV_UNDISTORT_MAP_INITIALIZER;
	float zoom = 0.6; // good for fisheye
	ret = mcv_init_undistort_map_file(&map, intrinsics_file, zoom);
	if(ret) return -1;

	// run dewarp process and average over number of loops
	int64_t t0 = _apps_time_monotonic_ns();
	for(int i=0;i<LOOPS;i++){
		ret = mcv_undistort_image(img_mat.data, output, &map);
		if(ret) return -1;
	}
	int64_t t1 = _apps_time_monotonic_ns();
	int64_t t_per_loop_ns = (t1-t0)/LOOPS;
	printf("%6.2fms   CPU undistort time\n", ((double)(t_per_loop_ns))/1000000.0);

	// cleanup
	mcv_undistort_map_clean(&map);
	return 0;
}

/**
 * @brief Does LOOPS number of iterations to determine the speed of the warping process on CVP 
 * 
 * @param img_mat The CV matrix to warp
 * @param intrinsics_file String location of the intrinsics file for the camera
 * @param output A buffer for the undistorted image
 * @return 0 if successful, -1 otherwise
 */
static int run_test_cvp(Mat img_mat, const char* intrinsics_file, uint8_t* output)
{
	int w = img_mat.size().width;
	int h = img_mat.size().height;

	// configure undistort parameters
	float zoom = 0.6;
	mcv_undistort_config_t config;
	config.width          = w;
	config.height         = h;
	config.warp_mode      = MCV_WARP_GRID;
	config.border_mode    = MCV_BORDER_CONSTANT;
	config.image_mode     = IMAGE_FORMAT_RAW8;
	config.undistort_grid = mcv_cvp_create_undistort_grid(w, h);

	// populate configurations undistortion grid for mono lens
	mcv_lens_params_t lens_params;
	mcv_create_lens_undistort_grid(intrinsics_file, zoom, &lens_params, config.undistort_grid);

	// use configuration to get an undistort handle
	mcv_cvp_undistort_handle undistort_h = mcv_cvp_undistort_init(config);
	if(undistort_h == NULL)
	{
		printf("Error in configuring undistort handle\n");
		return -1;
	}

	// run undistort process and average over number of loops
	int64_t t0 = _apps_time_monotonic_ns();
	for(int i=0;i<LOOPS;i++){
		if(mcv_cvp_undistort_process(undistort_h, img_mat.data, output))
		{
			printf("Error in undistortion process\n");
			return -1;
		}
	}
	int64_t t1 = _apps_time_monotonic_ns();
	int64_t t_per_loop_ns = (t1-t0)/LOOPS;
	printf("%6.2fms   CVP undistort time\n", ((double)(t_per_loop_ns))/1000000.0);

	// clean up undistortion
	if(mcv_cvp_undistort_deinit(undistort_h))
	{
		printf("Error: Failed to deinit undistort handle\n");
		return -1;
	}
	return 0;
}

/**
 * @brief Loads an image and tests undistortion times hardware
 * 
 * This is the primary unit test for this file. For a given image and
 * intrinsics file, this function computes the average time to undistort the 
 * image. If use_cvp is set to 0, this method uses the CPU on the VOXL to 
 * compute the undistortion. Otherwise, this method uses the less powerful 
 * DSPs to compute the undistortion
 * 
 * @param image_file String path to the image file for loading
 * @param output_file String path to the file to save the output
 * @param intrinsics_file String path to an intrinsics file for the camera
 * @param use_cvp Boolean flag to use CVP hardware
 * @return 0 if successful, -1 otherwise
 */
static int run_test(
	const char* image_file,
	const char* output_file, 
	const char* intrinsics_file, 
	int use_cvp
)
{
	int ret;

	// make output directory to save debug images to
	if(mkdir(OUTPUT_FOLDER, S_IRWXU) && errno!=EEXIST){
		fprintf(stderr, "failed to make folder\n");
		return -1;
	}

	// load in image using OpenCV
	Mat img_mat;
	img_mat = imread(image_file, IMREAD_GRAYSCALE);
	if(img_mat.type()!=CV_8UC1){
		fprintf(stderr, "ERROR loading %s\n", image_file);
		return -1;
	}

	// allocate for an output 8UC grayscale image
	int w = img_mat.size().width;
	int h = img_mat.size().height;
	uint8_t* output = (uint8_t*)malloc(w*h);

	// run the warp function using the hardware of choice
	if(use_cvp)
	{
		run_test_cvp(img_mat, intrinsics_file, output);
	} else {
		run_test_cpu(img_mat, intrinsics_file, output);	
	}

	// write out the output
	Mat output_mat(h, w, CV_8UC1, output);
	ret = imwrite(output_file, output_mat);
	if(!ret)
	{
		fprintf(stderr, "ERROR writing to %s\n", output_file);
	}

	// cleanup
	free(output);

	// printf("setting permissions for output files\n");
	ret = system("chmod 777 -R /data/libmodalcv/");
	if(ret){
		fprintf(stderr, "Warning, failed to set permissions for output files\n");
	}

	return 0;
}


// call run_test() on each test for both CVP and CPU
int main()
{
	// enable debug and timing printouts
	mcv_en_debug_prints(1);
	mcv_en_timing_prints(0);

	printf("--- TEST 1 ---\n");
	run_test(TEST_IMAGE_1, OUTPUT_FILE_1_CPU, INTRINSICS_FILE_1, false);
	run_test(TEST_IMAGE_1, OUTPUT_FILE_1_CVP, INTRINSICS_FILE_1, true);

	printf("--- TEST 2 ---\n");
	run_test(TEST_IMAGE_2, OUTPUT_FILE_2_CPU, INTRINSICS_FILE_2, false);
	run_test(TEST_IMAGE_2, OUTPUT_FILE_2_CVP, INTRINSICS_FILE_2, true);
	return 0;
}
